import copy
import logging
import sys
import yaml
import datetime
import qtawesome
from pydantic.color import Color as pydColor
from PyQt5 import QtWidgets, QtCore, QtGui
from redvypr.device import RedvyprDevice, RedvyprDeviceParameter
from redvypr.widgets.redvyprSubscribeWidget import redvyprSubscribeWidget
#from redvypr.widgets.gui_config_widgets import redvypr_ip_widget, configQTreeWidget, configWidget,
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget, pydanticDeviceConfigWidget, dictQTreeWidget, datastreamMetadataWidget
from redvypr.widgets.redvyprAddressWidget import datastreamWidget, datastreamsWidget
from redvypr.redvypr_address import RedvyprAddress
from redvypr.data_packets import RedvyprMetadata, RedvyprDeviceMetadata
import redvypr.files as files
import redvypr.device as device

_logo_file = files.logo_file
_icon_file = files.icon_file


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('gui')
logger.setLevel(logging.INFO)

# Some standard iconnames
iconnames = {'settings':'ri.settings-5-line'}


class deviceTableWidget(QtWidgets.QTableWidget):
    """
    Tablewidget that displays all devices of the redvypr instance, gives informations and controls possibility.

    """

    def __init__(self, redvyprWidget=None):
        """

        Args:
            redvypr:
            device:
        """
        super(deviceTableWidget, self).__init__()
        self.redvyprWidget = redvyprWidget
        self.redvypr = redvyprWidget.redvypr
        self.redvypr.device_added.connect(self.populate_table)
        self.redvypr.device_removed.connect(self.populate_table)
        self.populate_table()
        self._updatetimer = QtCore.QTimer()
        self._updatetimer.timeout.connect(self.updateDevicePackets)
        self._updatetimer.timeout.connect(self.deviceThreadStatusCheckAll)
        self._updatetimer.start(2000)
    def populate_table(self):
        nRows = len(self.redvypr.devices)
        colheader = ['Name', 'Start', 'Subscribe', 'Loglevel', 'Window location', 'Configure', 'View',
                         'Packets published', 'Packets received']

        self.colheader = colheader
        nCols = len(colheader)
        self.__startbuttons = []
        self.clear()
        self.setRowCount(nRows)
        self.setColumnCount(nCols)
        self.setHorizontalHeaderLabels(colheader)
        for irow,d in enumerate(self.redvypr.devices):
            #print('d',d)
            device = d['device']
            devicedict = d
            # Devicename
            colindex = colheader.index('Name')
            devicename = device.name
            item_name = QtWidgets.QTableWidgetItem(devicename)
            self.setItem(irow, colindex, item_name)
            # Start
            button_start = QtWidgets.QPushButton('Start')
            self.__startbuttons.append(button_start)
            button_start.setCheckable(True)
            button_start.__device = device
            device.thread_started.connect(self.deviceThreadStatusChanged)
            device.thread_stopped.connect(self.deviceThreadStatusChanged)
            button_start.__devicedict = devicedict
            button_start.clicked.connect(self.deviceStartStopClicked)
            colindex = colheader.index('Start')
            self.setCellWidget(irow, colindex, button_start)
            # Subscribe
            button_subscribe = QtWidgets.QPushButton('Subscribe')
            button_subscribe.__device = device
            button_subscribe.__devicedict = devicedict
            button_subscribe.clicked.connect(self.deviceSubscribeClicked)
            colindex = colheader.index('Subscribe')
            self.setCellWidget(irow, colindex, button_subscribe)
            # Loglevel
            if True:
                level = device.logger.getEffectiveLevel()
                levelname = logging.getLevelName(level)
                loglevels = ['INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL']
                log_combo = QtWidgets.QComboBox()
                for i, l in enumerate(loglevels):
                    log_combo.addItem(l)

                log_combo.setCurrentText(levelname)
                log_combo.__device = device
                log_combo.__devicedict = devicedict
                log_combo.currentIndexChanged.connect(self.loglevelChanged)
                colindex = colheader.index('Loglevel')
                self.setCellWidget(irow, colindex, log_combo)
            # Dock location
            combo_choose = QtWidgets.QComboBox()
            combo_choose.addItem('Tab')
            combo_choose.addItem('Window')
            combo_choose.addItem('Hide')
            combo_choose.__device = device
            combo_choose.__devicedict = devicedict
            # Change the combo to the device setup
            device_dockstr = device.device_parameter.gui_dock
            combo_choose.setCurrentText(device_dockstr)
            combo_choose.currentIndexChanged.connect(self.widgetlocChanged)

            colindex = colheader.index('Window location')
            self.setCellWidget(irow, colindex, combo_choose)
            # Configure
            button_configure = QtWidgets.QPushButton('Configure')
            button_configure.__device = device
            button_configure.__devicedict = devicedict
            button_configure.setEnabled(True)
            button_configure.clicked.connect(self.deviceConfigureClicked)
            colindex = colheader.index('Configure')
            self.setCellWidget(irow, colindex, button_configure)
            # View
            button_view = QtWidgets.QPushButton('View')
            button_view.__device = device
            button_view.__devicedict = devicedict
            button_view.clicked.connect(self.deviceViewClicked)
            colindex = colheader.index('View')
            self.setCellWidget(irow, colindex, button_view)

        self.resizeColumnsToContents()
        self.updateDevicePackets()

    def updateDevicePackets(self):
        """
        Updates the packets numbers in the table
        :return:
        """
        funcname = __name__ + '.updateDevicePackets():'
        #logger.debug(funcname)
        for irow, d in enumerate(self.redvypr.devices):
            npub = d['device'].statistics['packets_published']
            colindex = self.colheader.index('Packets published')
            item_pub = QtWidgets.QTableWidgetItem(str(npub))
            self.setItem(irow, colindex, item_pub)
            nrecv = d['device'].statistics['packets_received']
            colindex = self.colheader.index('Packets received')
            item_recv = QtWidgets.QTableWidgetItem(str(nrecv))
            self.setItem(irow, colindex, item_recv)

    def deviceViewClicked(self):
        funcname = __name__ + '.deviceViewClicked():'
        logger.debug(funcname)
        devicedict = self.sender().__devicedict
        tabindex = self.redvyprWidget.devicetabs.indexOf(devicedict['widget'])
        try:
            if tabindex > -1:
                self.redvyprWidget.devicetabs.setCurrentWidget(devicedict['widget'])
            else:
                devicedict['widget'].setFocus()
                devicedict['widget'].raise_()
                devicedict['widget'].activateWindow()
        except:
            logger.debug('View', exc_info=True)

    def deviceThreadStatusCheckAll(self):
        """
        Function checks the thread status of all devices and updates the buttons
        """
        for b in self.__startbuttons:
            device_tmp = b.__device
            startbutton = b
            status = device_tmp.get_thread_status()
            thread_status = status['thread_running']
            self.__update_start_button(startbutton, thread_status)
    def deviceThreadStatusChanged(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """
        device_tmp = self.sender()
        startbutton = None
        # Search for the correct startbutton
        for b in self.__startbuttons:
            if b.__device == device_tmp:
                startbutton = b
        # startbutton = self.__startbutton_clicked
        #print('Update!',device)
        status = device_tmp.get_thread_status()
        thread_status = status['thread_running']
        self.__update_start_button(startbutton, thread_status)

    def __update_start_button(self,startbutton, thread_status):
        if startbutton is not None:
            #print('Hall', status)
            #print('Hall', thread_status)
            # Running
            if (thread_status):
                startbutton.setText('Stop')
                startbutton.setChecked(True)
            # Not running
            else:
                startbutton.setText('Start')
                # Check if an error occured and the startbutton
                if (startbutton.isChecked()):
                    startbutton.setChecked(False)
                # self.conbtn.setEnabled(True)


    def deviceSubscribeClicked(self):
        funcname = __name__ + '.deviceSubscribeClicked()'
        button = self.sender()
        device = button.__device
        logger.debug(funcname + ':' + str(device))
        # self.__con_widget = redvyprConnectWidget(devices=self.redvypr.devices, device=device)
        self.__subscribeWidget = redvyprSubscribeWidget(redvypr=self.redvypr, device=device)
        self.__subscribeWidget.show()

    def deviceConfigureClicked(self):
        button = self.sender()
        funcname = __name__ + '.deviceConfigureClicked():'
        button = self.sender()
        device = button.__device
        logger.debug(funcname + ':' + str(device))
        self.__info_widget = redvypr_deviceInfoWidget(device)
        self.__info_widget.show()

    def deviceStartStopClicked(self):
        button = self.sender()
        #self.__startbutton_clicked = button
        device = button.__device
        if button.isChecked():
            logger.debug("button pressed")
            button.setText('Starting')
            device.thread_start()
            # self.device_start.emit(self.device)
        else:
            logger.debug('button released')
            button.setText('Stopping')
            button.setChecked(True)
            device.thread_stop()

    def loglevelChanged(self):
        devicedict = self.sender().__devicedict
        loglevel = self.sender().currentText()
        logger = devicedict['logger']
        logger.info('loglevel changed to {}'.format(loglevel))
        if(logger is not None):
            logger.setLevel(loglevel)


    def widgetlocChanged(self, index):
        """
        Functions checks if a device widget shall be attached to the tab layout
        or is a free floating window
        :param index:
        :return:
        """
        combo = self.sender()
        widgetloc = combo.currentText()
        devicewidget = combo.__devicedict['widget']
        device = combo.__device
        tabindex = self.redvyprWidget.devicetabs.indexOf(devicewidget)
        #print('Index', index, combo.__device, widgetloc, tabindex)
        widgetname = combo.__device.name
        if widgetloc == 'Window':
            if tabindex > -1:
                self.redvyprWidget.devicetabs.removeTab(tabindex)
                devicewidget.setParent(None)
                devicewidget.setWindowTitle(widgetname)
            device.device_parameter.gui_dock = widgetloc
            devicewidget.show()

        elif widgetloc == 'Hide':
            if tabindex > -1:
                self.redvyprWidget.devicetabs.removeTab(tabindex)
                devicewidget.setParent(None)
                devicewidget.setWindowTitle(widgetname)
            device.device_parameter.gui_dock = widgetloc
            devicewidget.show()
            devicewidget.hide()
        else:
            if tabindex == -1: #Add to tab, if not already there
                self.redvyprWidget.devicetabs.addTab(devicewidget, widgetname)
                device.device_parameter.gui_dock = widgetloc


class deviceTabWidget(QtWidgets.QTabWidget):
    def resizeEvent(self, event):
        #print("Window has been resized",event)
        #print('fds',event.size().width())
        wtran = event.size().width()-500
        #print('fsfsd',self.widget(1).width())
        self.setStyleSheet("QTabBar::tab:disabled {"+\
                        "width: {:d}px;".format(wtran)+\
                        "color: transparent;"+\
                        "background: transparent;}")
        super(deviceTabWidget, self).resizeEvent(event)



def get_QColor(data):
    """
    Returns a qcolor based on the data input, data can be either a string, a list of rgb or a dictionary of type {'r':250,'g':100,'b':0}
    """
    funcname = __name__ + '.get_QColor():'
    logger.debug(funcname)
    colordata = copy.deepcopy(data)
    #print('Colordata',colordata)
    #print('Type colordata', type(colordata))
    if(type(colordata) == str):
        color = QtGui.QColor(colordata)
    elif (type(colordata) == tuple):
        color = QtGui.QColor(colordata[0], colordata[1], colordata[2])
    elif (type(colordata) == list):
        color = QtGui.QColor(colordata[0], colordata[1], colordata[2])
    elif (type(colordata) == pydColor):
        colors = colordata.as_rgb_tuple()
        color = QtGui.QColor(colors[0], colors[1], colors[2])
    else:
        colors = colordata
        color = QtGui.QColor(colors['r'], colors['g'], colors['b'])

    return color

class redvyprAddDeviceWidget(QtWidgets.QWidget):
    """ A widget that lists all devices found in modules and in the python files included in the path list.

    """
    def __init__(self, redvypr_device_scan=None, redvypr=None):
        """

        Args:
            redvypr:
            device:
        """
        super(redvyprAddDeviceWidget, self).__init__()
        self.redvypr = redvypr
        if redvypr is not None:
            self.redvypr_device_scan = redvypr.redvypr_device_scan
        elif (redvypr_device_scan is not None):
            self.redvypr_device_scan = redvypr_device_scan
        else:
            self.redvypr_device_scan = device.RedvyprDeviceScan()

        # Update the devicetree
        self.create_tree_widget()
        self.update_tree_widget()
        self.create_deviceinfo_widget()
        # Create widgets for adding/removing devices
        self.addbtn = QtWidgets.QPushButton('Add')
        self.addbtn.clicked.connect(self.add_device_click)
        self.devnamelabel = QtWidgets.QLabel('Devicename')
        self.devname = QtWidgets.QLineEdit()
        self.mp_label = QtWidgets.QLabel('Multiprocessing options')
        self.mp_qthread = QtWidgets.QRadioButton('QThread')
        self.mp_thread = QtWidgets.QRadioButton('Thread')
        self.mp_multi = QtWidgets.QRadioButton('Multiprocessing')
        self.mp_group = QtWidgets.QButtonGroup()
        self.mp_group.addButton(self.mp_qthread)
        self.mp_group.addButton(self.mp_thread)
        self.mp_group.addButton(self.mp_multi)
        self.mp_qthread.setChecked(True)

        self.log_label = QtWidgets.QLabel('Loglevel')
        self.logwidget = QtWidgets.QComboBox()  # A Combobox to change the loglevel of the device
        # Fill the logwidget
        if (logger is not None):
            level = logger.getEffectiveLevel()
            levelname = logging.getLevelName(level)
            loglevels = ['INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL']
            for i, l in enumerate(loglevels):
                self.logwidget.addItem(l)

            self.logwidget.setCurrentText(levelname)
        else:
            self.logwidget.addItem('NA')

        thread_layout = QtWidgets.QHBoxLayout()
        thread_layout.addWidget(self.mp_qthread)
        thread_layout.addWidget(self.mp_multi)
        thread_layout.addWidget(self.mp_thread)
        self.layout = QtWidgets.QFormLayout(self)
        self.layout.addRow(self.devicetree)
        self.layout.addRow(self.deviceinfo)
        self.layout.addRow(self.log_label,self.logwidget)
        self.layout.addRow(self.mp_label,thread_layout)
        #self.layout.addRow(self.mp_thread, self.mp_multi)
        self.layout.addRow(self.devnamelabel, self.devname)
        self.layout.addRow(self.addbtn)

        self.setWindowIcon(QtGui.QIcon(_icon_file))
        self.setWindowTitle("redvypr add device")

    def create_deviceinfo_widget(self):
        """
        Creates a widget that shows the device information
        Returns:

        """
        self.deviceinfo = QtWidgets.QWidget()  #
        self.deviceinfo_layout = QtWidgets.QFormLayout(self.deviceinfo)
        self.__devices_info_sourcelabel2 = QtWidgets.QLabel()
        self.__devices_info_sourcelabel4 = QtWidgets.QLabel()
        self.__devices_info_sourcelabel6 = QtWidgets.QLabel()
        self.deviceinfo_layout.addRow(QtWidgets.QLabel('Name'),self.__devices_info_sourcelabel2)
        self.deviceinfo_layout.addRow(QtWidgets.QLabel('Source'),self.__devices_info_sourcelabel4)
        self.deviceinfo_layout.addRow(QtWidgets.QLabel('Description'),self.__devices_info_sourcelabel6)

    def create_tree_widget(self):
        """
        Creates the QtreeWidget with the
        Returns:

        """
        self.devicetree = QtWidgets.QTreeWidget()  # All dataproviding devices
        self.devicetree.setColumnCount(1)
        #self.devicetree.setHeaderHidden(True)
        self.devicetree.setHeaderLabels(['Device'])
        self.devicetree.currentItemChanged.connect(self.__item_changed__)
        self.devicetree.itemDoubleClicked.connect(self.__apply_item__)
        self.devicetree.setSortingEnabled(True)

    def update_tree_widget(self):
        self.devicetree.clear()
        root = self.devicetree.invisibleRootItem()
        #moduleroot = QtWidgets.QTreeWidgetItem(['modules', ''])
        #root.addChild(moduleroot)
        def update_recursive(moddict,parentitem):
            try:
                keys = moddict.keys()
            except:
                keys = None

            if(keys is None):
                return
            else:
                for k in moddict.keys():
                    if(k == '__devices__'): # List of devices in the module
                        for devdict in moddict[k]:
                            devicename = devdict['name']
                            #print('devdict',devdict)
                            # remove trailing modules separated by '.'
                            devicename = devicename.split('.')[-1]
                            itm = QtWidgets.QTreeWidgetItem([devicename, ''])
                            itm.devdict = devdict # Add device information
                            parentitem.addChild(itm)
                    else:
                        # remove trailing modules separated by '.'
                        if '/' in k: # Check if its a path or a module file
                            ktxt = k
                        else:
                            ktxt = k.split('.')[-1]

                        itm = QtWidgets.QTreeWidgetItem([ktxt, ''])
                        itm.devdict = None # Not a device
                        parentitem.addChild(itm)
                        update_recursive(moddict[k],itm)

        #update_recursive(self.redvypr_device_scan.redvypr_devices['modules'],moduleroot)
        update_recursive(self.redvypr_device_scan.redvypr_devices, root)


        self.devicetree.expandAll()
        self.devicetree.resizeColumnToContents(0)
        self.devicetree.sortByColumn(0, QtCore.Qt.AscendingOrder)

    def __item_changed__(self, new, old):
        logger.debug('Item changed')
        if new.devdict is not None:
            self.__update_device_info__(new.devdict)
            self.addbtn.setEnabled(True)
        else:
            self.addbtn.setEnabled(False)

    def __apply_item__(self):
        logger.debug('Apply')

    def __update_device_info__(self,devdict):
        """ Populates the self.__devices_info widget with the info of the module
        """
        infotxt = devdict['name']
        self.__devices_info_sourcelabel2.setText(infotxt)
        infotxt2 = devdict['file']
        self.__devices_info_sourcelabel4.setText(infotxt2)
        try:
            desctxt = devdict['module'].description
        except Exception as e:
            desctxt = ''

        self.__devices_info_sourcelabel6.setText(desctxt)

    def __device_name(self):
        devicemodulename = self.__devices_list.currentItem().text()
        devicename = devicemodulename + '_{:d}'.format(self.redvypr.numdevice + 1)
        self.__devices_devname.setText(devicename)
        self.__device_info()

    def add_device_click(self):
        """ Adds the device
        """
        funcname = __name__ + 'add_device_click():'
        logger.debug(funcname)
        getSelected = self.devicetree.selectedItems()
        if getSelected:
            item = getSelected[0]

        if item.devdict is not None:
            devicemodulename = item.devdict['name']
            device_parameter = RedvyprDeviceParameter()
            if self.mp_thread.isChecked():
                device_parameter.multiprocess = 'thread'
            elif self.mp_qthread.isChecked():
                device_parameter.multiprocess = 'qthread'
            elif self.mp_multi.isChecked():
                device_parameter.multiprocess = 'multiprocessing'

            levelname = self.logwidget.currentText()
            device_parameter.loglevel = levelname
            devname = str(self.devname.text())
            if len(devname) > 0:
                device_parameter.name = devname

            logger.debug('devicemodulename {}'.format(devicemodulename))
            logger.debug('Adding device, config {}'.format(device_parameter))
            if self.redvypr is not None:
                self.redvypr.add_device(devicemodulename=devicemodulename, base_config=device_parameter)
            self.devname.clear()
            # Update the name
            #self.__device_name()
            # Closing
            self.close()
        else:
            logger.warning(funcname + 'Not a device')




#
#
#
# A logging handler for qplaintext
#
#
#
class QPlainTextEditLogger(logging.Handler):
    def __init__(self):
        super(QPlainTextEditLogger, self).__init__()

    def add_widget(self,widget):        
        self.widget = widget

    def emit(self, record):
        msg = self.format(record)
        self.widget.appendPlainText(msg)

    def write(self, m):
        pass   
    



#
#
# Widget shows the statistics of the device
#
#
class redvypr_deviceStatisticWidget(QtWidgets.QWidget):
    """
    Widgets shows the device statistic as text
    """
    def __init__(self, device = None, dt_update=1000):
        funcname = __name__ + '.__init__():'
        logger.debug(funcname)
        super().__init__()
        self.device = device
        self.layout = QtWidgets.QGridLayout(self)
        self.infowidget       = QtWidgets.QPlainTextEdit()
        self.infowidget.setReadOnly(True)
        self.layout.addWidget(self.infowidget,0,0)

        self.__update_info()
        # Todo, let the user choose for an update
        #self.updatetimer = QtCore.QTimer()
        #self.updatetimer.timeout.connect(self.__update_info)
        #self.updatetimer.start(dt_update)

    def __update_info(self):
        funcname = __name__ + '.__update_info():'
        prev_cursor = self.infowidget.textCursor()
        pos  = self.infowidget.verticalScrollBar().value()
        pos2 = self.infowidget.verticalScrollBar().value()
        self.infowidget.clear()
        sortstat = {}
        for i in sorted(self.device.statistics):
            sortstat[i]=self.device.statistics[i]

        sortstat['datakeys'] = sorted(sortstat['datakeys'])
        statstr = yaml.dump(sortstat)
        self.infowidget.insertPlainText(statstr + '\n')
        #self.infowidget.moveCursor(QtGui.QTextCursor.End)
        # cursor.setPosition(0)
        # self.text.setTextCursor(prev_cursor)
        if True:
            self.infowidget.verticalScrollBar().setValue(pos)

class redvypr_deviceInfoWidget(QtWidgets.QWidget):
    """
    Information widget of a device
    """
    connect = QtCore.pyqtSignal(
        RedvyprDevice)  # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self, device = None, dt_update = 1000):
        funcname = __name__ + '.__init__():'
        logger.debug(funcname)
        super().__init__()
        self.device = device
        self.setWindowTitle(device.name)
        self.layout = QtWidgets.QGridLayout(self)
        tstr = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.update_label = QtWidgets.QLabel('Last update {:s}'.format(tstr))
        self.packetRecv_label = QtWidgets.QLabel('Packets received {:d}'.format(0))
        self.packetPubl_label = QtWidgets.QLabel('Packets published {:d}'.format(0))
        self.publist_label = QtWidgets.QLabel('Publishes to')
        self.publist = QtWidgets.QListWidget()
        self.sublist_label = QtWidgets.QLabel('Subscribed devices')
        self.sublist = QtWidgets.QListWidget()
        #self.subBtn = QtWidgets.QPushButton('Subscribe')
        #self.subBtn.clicked.connect(self.connect_clicked)
        self.metadataBtn = QtWidgets.QPushButton('Edit Device Metadata')
        self.metadataBtn.clicked.connect(self.metadata_clicked)
        if self.device.publishes:
            self.ddBtn = QtWidgets.QPushButton('Edit Datastream Metadata')
            self.ddBtn.clicked.connect(self.data_devices_clicked)
        self.confBtn = QtWidgets.QPushButton('Configure')
        self.confBtn.clicked.connect(self.config_clicked)
        self.statBtn = QtWidgets.QPushButton('Statistics')
        self.statBtn.clicked.connect(self.statistics_clicked)
        self.layout.addWidget(self.update_label)
        self.layout.addWidget(self.packetRecv_label)
        self.layout.addWidget(self.packetPubl_label)
        self.layout.addWidget(self.sublist_label)
        self.layout.addWidget(self.sublist)
        self.layout.addWidget(self.publist_label)
        self.layout.addWidget(self.publist)
        self.layout.addWidget(self.statBtn)
        if self.device.publishes:
            self.layout.addWidget(self.ddBtn)
        self.layout.addWidget(self.confBtn)
        self.layout.addWidget(self.metadataBtn)

        self.updatetimer = QtCore.QTimer()
        self.updatetimer.timeout.connect(self.__update_info)
        self.updatetimer.start(dt_update)

    def data_devices_clicked(self):
        funcname = __name__ + '.data_devices_clicked():'
        try:
            self.data_device_widget.setParent(None)
        except:
            pass

        # Filter the addresses
        device_filter_address = RedvyprAddress(uuid=self.device.address.uuid,publisher=self.device.address.publisher)
        filter_include = [device_filter_address]
        #print('Filter include',filter_include)
        #self.data_device_widget = datastreamWidget(redvypr=self.device.redvypr,filter_include=filter_include)
        self.data_device_widget = datastreamMetadataWidget(redvypr=self.device.redvypr, device=self.device, filter_include=filter_include)
        self.data_device_widget.setWindowTitle(self.device.name)
        self.data_device_widget.show()

    def config_clicked(self):
        funcname = __name__ + '.config_clicked():'
        logger.debug(funcname)
        self.config_widget = pydanticDeviceConfigWidget(self.device)
        #targetWidget.showFullScreen()
        self.config_widget.showMaximized()

    def metadata_clicked(self):
        funcname = __name__ + '.metadata_clicked():'
        logger.debug(funcname)
        metadata_device = copy.deepcopy(self.device.statistics['metadata'])
        deviceAddress = RedvyprAddress(devicename=self.device.name)
        try:
            metadata_raw = metadata_device[deviceAddress.address_str]
        except:
            logger.info('Could not load metadata', exc_info=True)
            metadata_raw = {}

        metadata = RedvyprDeviceMetadata(**metadata_raw)
        print('Metadata',metadata)
        self.__metadata_edit = metadata
        self.__metadata_address = deviceAddress
        self.metadata_config = pydanticConfigWidget(metadata, configname=deviceAddress.address_str)
        self.metadata_config.config_editing_done.connect(self.metadata_config_apply)
        self.metadata_config.show()

    def metadata_config_apply(self):
        funcname = __name__ + '.metadata_config_apply():'
        logger.debug(funcname)

        print('Metadata new',self.__metadata_edit)
        metadata = self.__metadata_edit.model_dump()
        self.device.set_metadata(self.__metadata_address, metadata)

    def statistics_clicked(self):
        funcname = __name__ + '.statistics_clicked():'
        logger.debug(funcname)
        self.statistics_widget = redvypr_deviceStatisticWidget(device=self.device)
        self.statistics_widget.show()

    def subscribe_clicked(self):
        funcname = __name__ + '.subscribe_clicked():'
        logger.debug(funcname)
        button = self.sender()
        self.__subscribeWidget = redvyprSubscribeWidget(redvypr=self.redvypr, device=self.device)
        self.__subscribeWidget.show()
        #self.connect.emit(self.device)

    def __update_info(self):
        funcname = __name__ + '.__update_info():'
        #logger.debug(funcname)
        tstr = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.update_label.setText('Last update {:s}'.format(tstr))
        nrecv = self.device.statistics['packets_received']
        npub = self.device.statistics['packets_published']
        self.packetRecv_label.setText('Packets received {:d}'.format(nrecv))
        self.packetPubl_label.setText('Packets published {:d}'.format(npub))

        devs = self.device.get_subscribed_devices()
        self.sublist.clear()
        for d in devs:
            devname = d.name
            self.sublist.addItem(devname)

        devs_sub = self.device.publishing_to()
        self.publist.clear()
        for d in devs_sub:
            devname = d.name
            self.publist.addItem(devname)

# Can be removed soon
class deviceControlWidget_legacy(QtWidgets.QWidget):
    """ A widget to set the general settings of the device (start/stop, debug level)
    """
    device_start = QtCore.pyqtSignal(dict)  # Signal requesting a start of the device (starting the thread)
    device_stop = QtCore.pyqtSignal(dict)  # Signal requesting a stop of device
    connect = QtCore.pyqtSignal(dict)  # Signal requesting a change of the connection

    def __init__(self, devicedict, redvyprwidget):
        super(deviceControlWidget, self).__init__()
        self.devicedict = devicedict
        self.device = devicedict['device']
        # Connect the status signal
        self.device.status_signal.connect(self.thread_status)
        self.redvyprwidget = redvyprwidget
        self.devicetab = self.redvyprwidget.devicetabs  # The parent tab with all devices listed
        self.namelabel = QtWidgets.QLabel(devicedict['device'].name)
        label = self.namelabel
        fsize = label.fontMetrics().size(0, label.text())
        label.setFont(QtGui.QFont('Arial', fsize.height() + 4))
        # self.numlabel = QtWidgets.QLabel(str(devicedict['device'].numdevice))
        # label = self.numlabel
        # fsize         = label.fontMetrics().size(0, label.text())
        # label.setFont(QtGui.QFont('Arial', fsize.height()+4))
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout2 = QtWidgets.QGridLayout()
        self.logwidget = QtWidgets.QComboBox()  # A Combobox to change the loglevel of the device
        self.logwidget.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        # Fill the logwidget
        logger = self.devicedict['logger']
        if (logger is not None):
            level = logger.getEffectiveLevel()
            levelname = logging.getLevelName(level)
            loglevels = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']
            for i, l in enumerate(loglevels):
                self.logwidget.addItem(l)

            self.logwidget.setCurrentText(levelname)
        else:
            self.logwidget.addItem('NA')

        self.logwidget.currentIndexChanged.connect(self.loglevel_changed)

        self.viewbtn = QtWidgets.QPushButton("View")
        self.viewbtn.clicked.connect(self.viewclicked)
        self.viewbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.conbtn = QtWidgets.QPushButton("Subscribe")
        self.conbtn.clicked.connect(self.conclicked)
        self.conbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.rembtn = QtWidgets.QPushButton("Remove")
        self.rembtn.clicked.connect(self.remdevice)
        self.rembtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.renbtn = QtWidgets.QPushButton("Rename")
        self.renbtn.clicked.connect(self.rendevice)
        self.renbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.startbtn = QtWidgets.QPushButton("Start")
        self.startbtn.setCheckable(True)
        self.startbtn.clicked.connect(self.startstopclicked)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self.infobtn = QtWidgets.QPushButton("Info")
        self.infobtn.clicked.connect(self.get_info)
        self.infobtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        # Configuration
        configlayout = QtWidgets.QVBoxLayout()
        # Autostart
        self.autostart = QtWidgets.QPushButton('Autostart')
        self.autostart.setCheckable(True)
        self.autostart.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        try:  # If the device has an autostart attribute
            self.autostart.isChecked(self.device.autostart)
        except:
            pass
        self.autostart.clicked.connect(self.autostart_clicked)
        # self.layout2.addWidget(QtWidgets.QLabel('Name' + ' Device #' + str(devicedict['device'].numdevice)),0,0)
        self.layout2.addWidget(self.namelabel, 0, 0)
        self.layout.addLayout(self.layout2)
        self.layout.addStretch()
        self.layout.addWidget(self.autostart)
        self.layout.addWidget(self.logwidget)
        self.layout.addWidget(self.viewbtn)
        self.layout.addWidget(self.infobtn)
        self.layout.addWidget(self.renbtn)
        self.layout.addWidget(self.conbtn)
        self.layout.addWidget(self.rembtn)
        self.layout.addWidget(self.startbtn)

    def autostart_clicked(self):
        # print('Autostart',self.autostart.isChecked())
        self.device.autostart = self.autostart.isChecked()

    def loglevel_changed(self):
        loglevel = self.logwidget.currentText()
        logger = self.devicedict['logger']
        logger.info('loglevel changed to {}'.format(loglevel))
        if (logger is not None):
            logger.setLevel(loglevel)

    def get_info(self):
        self.infowidget = QtWidgets.QPlainTextEdit()
        self.infowidget.setReadOnly(True)
        sortstat = {}
        for i in sorted(self.devicedict['statistics']):
            sortstat[i] = self.devicedict['statistics'][i]

        sortstat['datakeys'] = sorted(sortstat['datakeys'])
        statstr = yaml.dump(sortstat)
        self.infowidget.insertPlainText(statstr + '\n')
        self.infowidget.show()

    def viewclicked(self):
        self.redvyprwidget.devicetabs.setCurrentWidget(self.devicedict['widget'])

    def conclicked(self):
        self.connect.emit(self.devicedict)

    def startstopclicked(self):
        funcname = __name__ + '.startstopclicked()'
        logger.debug(funcname)
        if (self.startbtn.text() == 'Stop'):
            self.device.thread_stop()
        else:
            self.device.thread_start()

    def thread_status(self, statusdict):
        """ Function regularly called by redvypr to update the thread status
        """
        status = statusdict['thread_running']
        if (status):
            self.startbtn.setText('Stop')
            self.startbtn.setChecked(True)
        else:
            self.startbtn.setText('Start')
            self.startbtn.setChecked(False)

    def remdevice(self):
        """ Removing the device
        """
        ret = QtWidgets.QMessageBox.question(self, '', "Are you sure to remove the device?",
                                             QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if ret == QtWidgets.QMessageBox.Yes:
            widget = self.devicedict['widget']
            for i in range(self.devicetab.count()):
                if (self.devicetab.widget(i) == widget):
                    self.redvyprwidget.closeTab(i)

    def rendevice(self):
        """ Renaming a device
        """
        oldname = self.devicedict['device'].name
        name, okPressed = QtWidgets.QInputDialog.getText(self, "Enter new name", "Device name:",
                                                         QtWidgets.QLineEdit.Normal, oldname)
        if okPressed and name != '':
            renamed = self.redvyprwidget.renamedevice(oldname, name)





import copy
import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import serial
import serial.tools.list_ports
import logging
import sys
import pydantic
import typing
import redvypr.data_packets
from redvypr.data_packets import check_for_command, create_datadict
from redvypr.device import RedvyprDevice
import redvypr.files as redvypr_files
import redvypr.devices.plot.plot_widgets as redvypr_plot_widgets
import redvypr.widgets.standard_device_widgets #import redvypr_deviceInitWidget, displayDeviceWidget_standard

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('example_device_with_gui')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True


def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    i = 0
    counter = 0
    while True:
        try:
            data = datainqueue.get(block = False)
        except:
            data = None
        if(data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            if (command == 'stop'):
                logger.debug('Got a command: {:s}'.format(str(data)))
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break

        data = redvypr.data_packets.create_datadict(device = device_info['device'])
        data['data'] = float(np.random.rand(1)-0.5)
        data['sometext'] = 'Hallo {}'.format(counter)
        dataqueue.put(data)
        time.sleep(1.0)


class initDeviceWidget(redvypr.widgets.standard_device_widgets.redvypr_deviceInitWidget):
    def __init__(self, *args, **kwargs):
        funcname = __name__ + '__init__():'
        logger.debug(funcname)
        super().__init__(*args, **kwargs)
class displayDeviceWidget(redvypr.widgets.standard_device_widgets.displayDeviceWidget_standard):
    def __init__(self, *args, **kwargs):
        funcname = __name__ + '__init__():'
        logger.debug(funcname)
        super().__init__(*args, **kwargs)



"""

redvypr device class



"""

import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import logging
import sys
import yaml
import copy
import uuid
import multiprocessing
import threading
import importlib
import glob
import pathlib
import inspect
import pkg_resources
import redvypr
import pydantic
import typing
import re
from redvypr.data_packets import commandpacket, create_datadict
from redvypr.packet_statistic import do_data_statistics
from redvypr.redvypr_address import RedvyprAddress

logging.basicConfig(stream=sys.stderr)


# The maximum size the dataqueues have, this should be more than
# enough for a "normal" usage case
queuesize = 10000
# queuesize = 10

class DeviceMetadata(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    description: str = pydantic.Field(default='')
    lon: float = pydantic.Field(default=-9999)
    lat: float = pydantic.Field(default=-9999)

class RedvyprDeviceCustomConfig(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    config_type: typing.Literal['custom'] = pydantic.Field(default='custom')

class RedvyprDeviceBaseConfig(pydantic.BaseModel):
    """
    This is the base config of any redvypr device.
    """
    name: str = pydantic.Field(default='')
    multiprocess: str = pydantic.Field(default='qthread')
    loglevel: str = pydantic.Field(default='')
    autostart: bool = False
    clear_datainqueue_before_thread_starts: bool = pydantic.Field(default=False, description='Clears the datainqueue before the thread is started.')
    devicemodulename: str = pydantic.Field(default='')
    description: str = ''
    gui_tablabel_init: str = 'Init'
    gui_tablabel_display: str = 'Display'
    gui_dock: typing.Literal['Tab','Window','Hide'] = pydantic.Field(default='Tab')
    gui_icon: str = 'mdi.network-outline'

class RedvyprDeviceConfig(pydantic.BaseModel):
    """
    Device configuration that is used for saving and loading files
    """
    base_config: RedvyprDeviceBaseConfig = pydantic.Field(default=RedvyprDeviceBaseConfig())
    devicemodulename: str = pydantic.Field(default='', description='')
    custom_config: typing.Optional[RedvyprDeviceCustomConfig] = pydantic.Field(default=None, description='')
    subscriptions: list = pydantic.Field(default=[])
    datakeytags: list = pydantic.Field(default=['mac','sn','id'],description='datakeys for which will be looked at at each packet and a statistic will be done')
    metadata: typing.Optional[DeviceMetadata] = pydantic.Field(default=None, description='')
    config_type: typing.Literal['device'] = pydantic.Field(default='device')

class RedvyprDeviceParameter(RedvyprDeviceBaseConfig):
    """
    This is the base config with extra parameter that dont need to be saved
    """
    uuid: str = pydantic.Field(default='')
    #template: dict = pydantic.Field(default={}) # Candidate for removal
    #config: dict = pydantic.Field(default={})  # Candidate for removal
    publishes: bool = True
    subscribes: bool = True
    numdevice: int = pydantic.Field(default=-1)
    # Not as parameter, but necessary for initialization
    maxdevices: int = pydantic.Field(default=-1)

class deviceQThread(QtCore.QThread):
    def __init__(self, startfunction, start_arguments):
        QtCore.QThread.__init__(self)
        self.startfunction = startfunction
        self.start_arguments = start_arguments

    def __del__(self):
        self.wait()

    def run(self):
        #print('Arguments',self.start_arguments)
        self.startfunction(*self.start_arguments)


def device_start_standard(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start():'
    logger = logging.getLogger('device_start():')
    while True:
        data = datainqueue.get()
        if(data is not None):
            command = redvypr.data_packets.check_for_command(data, thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(data)))
            if (command == 'stop'):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break

            #print('data',data)
            dataqueue.put(data)



logging.basicConfig(stream=sys.stderr)


class RedvyprDeviceScan():
    """
    Searches for redvypr devices
    """
    def __init__(self, device_path = [],
                 scan=True,
                 scan_redvypr=True,
                 scan_modules=True,
                 scan_devicepath=True,
                 redvypr_devices=None,
                 loglevel = logging.INFO):
        """

        Parameters
        ----------
        device_path
        scan: bool
            Does a scan after initialization.
        scan_redvypr: bool
            Scans redvypr_devices (if scan is set)
        scan_modules: bool
            Scans modules (if scan is set)
        scan_devicepath: bool
            Scans possible devices in devicepaths (if scan is set)
        redvypr_devices: python redvypr modules
            Scans modules for redvypr compatible devices
        loglevel: logging.loglevel
            The loglevel
        """
        self.logger = logging.getLogger('redvypr_device_scan')
        self.logger.setLevel(loglevel)
        self.device_paths = device_path
        self.redvypr_devices = redvypr_devices
        #self.redvypr = redvypr
        self.device_modules_path = []
        self.device_modules = []
        self.redvypr_devices = {'redvypr_modules': {},'redvypr':{},'files':{}}
        self.redvypr_devices_flat = []
        self.__modules_scanned__ = []
        self.__modules_scanned__.append(redvypr) # Do not scan redvypr itself

        # Start scanning
        if(scan):
            if scan_redvypr:
                if redvypr_devices is not None:
                    if isinstance(redvypr_devices,list):
                        for redvypr_device in redvypr_devices:
                            self.scan_redvypr(redvypr_device)
                    else:
                        self.scan_redvypr(redvypr_devices)

            if scan_modules:
                self.scan_modules()
            if scan_devicepath:
                self.scan_devicepath()


    def print_modules(self):
        for m in self.device_modules:
            print(m)


    def scan_devicepath(self):
        funcname = 'search_in_path():'
        self.logger.debug(funcname)
        self.device_modules = []  # Clear the list
        #
        # Add all devices from additionally folders
        #
        # https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported
        for dpath in self.device_paths:
            python_files = glob.glob(dpath + "/*.py")
            self.logger.debug(funcname + 'Will search in path for files: {:s}'.format(dpath))
            for pfile in python_files:
                self.logger.debug(funcname + 'Opening {:s}'.format(pfile))
                module_name = pathlib.Path(pfile).stem
                spec = importlib.util.spec_from_file_location(module_name, pfile)
                module = importlib.util.module_from_spec(spec)
                try:
                    spec.loader.exec_module(module)
                except:
                    self.logger.warning(funcname + 'Could not import module: {:s}\n--------------------------------------------\n'.format(pfile))
                    self.logger.warning('Because',exc_info=True)
                    self.logger.warning(funcname + '\n--------------------------------------------\n')

                module_members = inspect.getmembers(module, inspect.isclass)
                valid_module = self.valid_device(module)
                if (valid_module['valid']):  # If the module is valid add it to devices
                    devdict = {'module': module, 'name': module_name, 'file': module.__file__,'type':'file'}
                    # Test if the module is already there, otherwise append
                    if (module in self.__modules_scanned__):
                        # logger.debug(funcname + ': Module has been tested already ...')
                        continue
                    else:
                        try:
                            self.redvypr_devices['files'][pfile] = {'__devices__': [devdict]}
                        except Exception as e:
                            self.logger.exception(e)

                        self.redvypr_devices_flat.append(devdict)
                        self.__modules_scanned__.append(module)
                        self.logger.debug(funcname + 'Found device: {}'.format(module_name))
                else:
                    self.logger.debug(funcname + 'Not a valid device')



    def scan_module_recursive(self,testmodule, module_dict):
        funcname = 'scan_module_recursive():'
        self.logger.debug(funcname + ' Scanning: {}'.format(testmodule))
        #print(funcname,testmodule)
        # Check if the device is valid
        valid_module = self.valid_device(testmodule)
        #print('Valid dictionary',valid_module)
        if (valid_module['valid']):  # If the module is valid add it to devices
            # print('Members',inspect.getmembers(testmodule, inspect.ismodule))
            devdict = {'module': testmodule, 'name': testmodule.__name__, 'file': testmodule.__file__,'type':'module'}
            # module_dict[testmodule.__name__] = devdict
            try:
                module_dict['__devices__'].append(devdict)
            except:
                module_dict['__devices__'] = [devdict]

            self.redvypr_devices_flat.append(devdict)
            self.logger.debug(funcname + ' Found valid module {:s}'.format(str(testmodule)))
        # Always append as scanned
        self.__modules_scanned__.append(testmodule)

        # Checks if the module has a variable called redvypr_devicemodule
        if (valid_module['hasredvyprdevicemodule']):  # If the module is valid add it to devices
            device_module_tmp = inspect.getmembers(testmodule, inspect.ismodule)
            if len(device_module_tmp) > 0:
                for smod in device_module_tmp:
                    testmodule2 = getattr(testmodule, smod[0])
                    if (testmodule2 in self.__modules_scanned__):
                        # logger.debug(funcname + ': Module has been tested already ...')
                        continue
                    else:
                        try:
                            module_dict[testmodule.__name__]
                        except:
                            module_dict[testmodule.__name__] = {}
                        self.scan_module_recursive(testmodule2, module_dict[testmodule.__name__])
                        # Cleanup modules without devices
                        if len(module_dict[testmodule.__name__].keys()) == 0:
                            module_dict.pop(testmodule.__name__)

    def scan_redvypr(self,redvyprdevices):
        funcname = 'scan_redvypr():'
        self.logger.debug(funcname)
        if True:
            try:
                device_module_all = inspect.getmembers(redvyprdevices)
                self.scan_module_recursive(redvyprdevices,self.redvypr_devices['redvypr'])

            except Exception as e:
                self.logger.exception(e)
                #self.logger.info(funcname + ' Could not import module: ' + str(e))# If the module is valid add it to devices


    def scan_modules(self,package_names = ['vypr','vyper']):
        """
        Searches for installed packages containing redvypr devices. Packages do typically start with redvypr_package_name
        and add them to the device path
        Updates self.device_modules with dictionaries of type
        devdict = {'module': module, 'name': module_name, 'source': module.__file__}
        with module the imported module with module_name at the location __file__.

        Args:
            package_names: a list of package names that will be inspected
        Returns:

        """

        funcname = 'scan_modules():'
        self.logger.debug(funcname)

        # Loop over all modules and search for devices
        for d in pkg_resources.working_set:
            FLAG_POTENTIAL_MODULE = False
            #print('Package', d.location, d.project_name, d.version, d.key)
            #print(d.key)
            for name in package_names:
                if name in d.key:
                    FLAG_POTENTIAL_MODULE = True

            # Dont import the redvypr module itself
            if d.key == 'redvypr':
                FLAG_POTENTIAL_MODULE = False

            if(FLAG_POTENTIAL_MODULE):
                #print('Found potential package',d.location, d.project_name, d.version, d.key)
                libstr2 = d.key.replace('-','_')  # Need to replace - with _, because - is not allowed in python
                try:
                    testmodule = importlib.import_module(libstr2)
                    device_module_all = inspect.getmembers(testmodule)
                    #print('Scan recursive start')
                    self.scan_module_recursive(testmodule,self.redvypr_devices['redvypr_modules'])
                    # Clean empty dictionaries

                except Exception as e:
                    #self.logger.info(funcname + ' Could not import module: ' + str(e))  # If the module is valid add it to devices
                    self.logger.debug('Could not import module', exc_info=True)


    def valid_device(self, devicemodule):
        """ Checks if the module is a valid redvypr module
        """
        funcname = 'valid_device(): '
        #self.logger.debug(funcname + 'Checking device {:s}'.format(str(devicemodule)))
        try:
            devicemodule.config_template
            hastemplate = True
        except:
            hastemplate = False

        try:
            devicemodule.Device
            hasdevice = True
        except:
            hasdevice = False

        try:
            devicemodule.start
            hasstart = True
        except:
            hasstart = False

        if (hasstart == False):
            try:
                devicemodule.Device.start
                hasstart = True
            except:
                hasstart = False

        try:
            devicemodule.displayDeviceWidget
            hasdisplaywidget = True
        except:
            hasdisplaywidget = False

        try:
            devicemodule.initDeviceWidget
            hasinitwidget = True
        except:
            hasinitwidget = False

        try:
            devicemodule.redvypr_devicemodule
            hasredvyprdevicemodule = True
        except:
            hasredvyprdevicemodule = False

        devicecheck = {}
        devicecheck['valid'] = hasstart and hasredvyprdevicemodule
        devicecheck['hasdevice'] = hasdevice
        devicecheck['hasredvyprdevicemodule'] = hasredvyprdevicemodule
        devicecheck['hastemplate'] = hastemplate
        devicecheck['start'] = hasstart
        devicecheck['initgui'] = hasinitwidget
        devicecheck['displaygui'] = hasdisplaywidget

        return devicecheck


# TODO: properly implement status signal with status dict similar to thread_started/stopped
class RedvyprDevice(QtCore.QObject):
    thread_started = QtCore.pyqtSignal(dict)  # Signal notifying that the thread started
    thread_stopped = QtCore.pyqtSignal(dict)  # Signal notifying that the thread started
    status_signal  = QtCore.pyqtSignal(dict)   # Signal with the status of the device
    subscription_changed_signal = QtCore.pyqtSignal()  # Signal notifying that a subscription changed
    config_changed_signal = QtCore.pyqtSignal()  # Signal notifying that the configuration of the device has changed

    def __init__(self, device_parameter = None, redvypr=None, dataqueue=None, comqueue=None, datainqueue=None,
                 statusqueue=None, guiqueues=None, custom_config=None, statistics=None, startfunction=None):
        """
        """
        super(RedvyprDevice, self).__init__()
        self.publishes = device_parameter.publishes    # publishes data, a typical sensor is doing this
        self.subscribes = device_parameter.subscribes   # subscribes other devices data, a typical datalogger is doing this
        self.datainqueue = datainqueue
        self.dataqueue = dataqueue
        self.comqueue = comqueue
        self.statusqueue = statusqueue
        self.custom_config = custom_config
        self.distribute_data_replyqueue = queue.Queue(maxsize=500)
        self.redvypr = redvypr
        self.name = device_parameter.name
        self.devicemodulename = device_parameter.devicemodulename
        self.uuid = device_parameter.uuid
        self.guiqueues = guiqueues
        try:
            self.host_uuid = redvypr.hostinfo['uuid']
        except:
            self.host_uuid = ''
        self.thread_uuid = ''
        self.host = redvypr.hostinfo
        self.loglevel = device_parameter.loglevel
        self.numdevice = device_parameter.numdevice
        self.description = 'redvypr_device'
        self.statistics = statistics
        self.mp = device_parameter.multiprocess
        self.autostart = device_parameter.autostart
        self.thread = None
        self.device_parameter = device_parameter
        # Create a redvypr_address
        # self.address_str
        # self.address
        self.__update_address__()

        # Add myself to the statistics
        datapacket = create_datadict(device=self.name,
                                     packetid=self.name,
                                     hostinfo=self.host)
        datapacket['_redvypr']['devicemodulename'] = self.devicemodulename
        do_data_statistics(datapacket, self.statistics)

        # Adding the start function (the function that is executed as a thread or multiprocess and is doing all the work!)
        if(startfunction is not None):
            self.start = startfunction

        # The queue that is used for the thread commmunication
        self.thread_communication = self.datainqueue

        self.subscribed_addresses = []
        
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(device_parameter.loglevel)
        # Some placeholder attribute, that will be filled by redvypr_main_widget, if the gui is used
        self.deviceinitwidget = None
        self.devicedisplaywidget = None
        self.redvyprdevicewidget = None
        # Timer to ping the status of thread
        self.__stoptimer__ = QtCore.QTimer()
        self.__stoptimer__.timeout.connect(self.__check_thread_status)  # Add to the timer another update

    def add_guiqueue(self, widget=None):
        """
        Adds a guiqueue to the guiqueues list. The queue can be used internally to process and display the data that is
        sent, additionally also a widget can be added. The widget needs to provide a function .update_data.
        :param widget:
        :return: (Queue, widget)
        """
        if 'thread' in self.device_parameter.multiprocess:  # Thread or QThread
            guiqueue = queue.Queue(maxsize=queuesize)
        else: # multiprocess
            guiqueue = multiprocessing.Queue(maxsize=queuesize)

        newqueue = (guiqueue,widget)
        self.guiqueues.append(newqueue)
        return newqueue

    def __update_address__(self):
        # self.address_str = self.name + ':' + self.redvypr.hostinfo['hostname'] + '@' + self.redvypr.hostinfo[
        #    'addr'] + '::' + self.redvypr.hostinfo['uuid']
        # self.address = redvypr_address(self.address_str)
        self.address = RedvyprAddress(devicename=self.name, local_hostinfo=self.redvypr.hostinfo, publisher=self.name)
        self.address_str = self.address.get_str(address_format='/u/a/h/p/d/')

    def config_changed(self):
        """
        Function should be called when the configuration of the device has changed
        :return:
        """
        self.config_changed_signal.emit()
        # TODO: Status changed to be emmited as well

    def subscription_changed_global(self, devchange):
        """
        Function is called by redvypr after another device emitted the subscription_changed_signal

        Args:
            devchange: redvypr_device that emitted the subscription changed signal

        Returns:

        """
        pass
        #print('Global subscription changed',self.name,devchange.name)

    def address_from_datakey(self,datakey):
        """
        Returns a datastream string from the datakey and self.address
        Args:
            key:

        Returns:
            str: datastream strong
        """
        addr = RedvyprAddress(datakey = datakey, devicename=self.name, local_hostinfo=self.redvypr.hostinfo)
        return addr

    def subscribe_address(self, address, force=False):
        """
        Subscribes to address
        Args:
            address:
            force:

        Returns:

        """
        funcname = self.__class__.__name__ + '.subscribe_address()'
        self.logger.debug(funcname + ' subscribing to device {:s}'.format(str(address)))
        #print('Address',address,type(address))
        if type(address) == str:
            raddr = RedvyprAddress(str(address))
        elif type(address) == RedvyprAddress:
            raddr = address
        else:
            raise TypeError('address needs to be a str or a redvypr_address')

        FLAG_NEW = True
        # Test if the same address exists already
        for a in self.subscribed_addresses:
            if(a.address_str == raddr.address_str):
                FLAG_NEW = False
                break

        if(FLAG_NEW):
            self.subscribed_addresses.append(raddr)
            self.subscription_changed_signal.emit()
            #print(self.subscribed_addresses)
            return True
        else:
            if force: # Resend the subscription signal
                self.subscription_changed_signal.emit()
            return False

    def unsubscribe_address(self, address):
        """
        """
        funcname = self.__class__.__name__ + '.unsubscribe_address()'
        self.logger.debug(funcname + ' unsubscribing from device {:s}'.format(str(address)))
        #print('Address', address, type(address))
        if (type(address) == str):
            raddr = RedvyprAddress(address)
        else:
            raddr = address

        try:
            self.subscribed_addresses.remove(raddr)
            self.subscription_changed_signal.emit()
        except Exception as e:
            self.logger.warning('Could not remove address {:s}: {:s}'.format(str(address),str(e)))

    def unsubscribe_all(self):
        """
        """
        funcname = self.__class__.__name__ + '.unsubscribe_all():'
        self.logger.debug(funcname + ' unsubscribing all')
        while len(self.subscribed_addresses)>0:
            try:
                self.subscribed_addresses.pop(0)
            except Exception as e:
                self.logger.warning(funcname + 'Could not remove address')

        self.subscription_changed_signal.emit()


    def change_name(self,name):
        """
        Changes the name of the device

        Args:
            name:

        Returns:

        """
        self.name = name
        self.__update_address__()
        # Clear the statistics
        self.statistics['numpackets'] = 0
        self.statistics['datakeys'] = []
        self.statistics['devicekeys'] = {}
        self.statistics['devices'] = []
        self.statistics['datastreams'] = []
        self.statistics['datastreams_dict'] = {}
        self.statistics['datastreams_info'] = {}
        self.logger.warning('This changes only the device name but will not restart the thread.')

    def address_string(self, address_format='/u/a/h/p/d'):
        """
        Returns the address string of the device
        Returns:

        """
        astr = self.address.get_str(address_format = address_format)
        return astr

    def got_subscribed(self, dataprovider_address, datareceiver_address):
        """
        Function is called by self.redvypr if this device is connected with another one. The intention is to notify device
        that a specific datastream/devce has been subscribed. This is a wrapper function and need to be reimplemented if needed.
        See i.e. iored as an example
        Args:
            dataprovider_address:
            datareceiver_address:

        Returns:

        """
        pass

    def got_unsubscribed(self, dataprovider_address, datareceiver_address):
        """

        Args:
            dataprovider_address:
            datareceiver_address:

        Returns:

        """
        pass


    def isRunning(self):
        info_dict = self.get_thread_status()
        return info_dict['thread_running']

    def get_thread_status(self):
        """

        Returns:

        """
        running = False
        running = self.thread_running()
        info_dict = {}
        info_dict['uuid'] = self.uuid
        info_dict['thread_uuid'] = self.thread_uuid
        info_dict['thread_running'] = running

        return info_dict


    def __send_command__(self,command):
        """
        Sends a command to the running thread, either via the comqueue or the datainqueue
        Args:
            command: The command to be sent to the device, typically created with datapacket.commandpacket

        Returns:

        """

        self.thread_communication.put(command)


    def kill_process(self):
        funcname = __name__ + '.kill_process():'
        self.logger.debug(funcname)
        if self.thread is not None:
            if isinstance(self.thread, deviceQThread):
                self.logger.debug(funcname + ': Terminating now')
                self.thread.terminate()
            #elif(self.mp == 'multiprocess'):
            elif isinstance(self.thread, deviceQThread):
                self.logger.debug(funcname + ': Terminating now')
                self.thread.kill()
            elif isinstance(self.thread, threading.Thread):
                self.logger.debug(funcname + ': Cannot terminate a threading.Thread')
            else:
                self.logger.warning(funcname + ' Thread type no known')
        else:
            self.logger.warning(funcname + ' Device is not running')

    def thread_running(self):
        try: # thread/mulitprocess
            running = self.thread.is_alive()
        except:  # QThread
            try:
                running = self.thread.isRunning()
            except:
                running = False

        return running

    def thread_command(self, command, data=None):
        """
        Sends a command to the device thread
        Args:
            command: string, i.e. "stop"
            data: dictionary with additional data, the data will be incorporated into the command dict by executing command.update(data)

        Returns:

        """
        funcname = __name__ + '.thread_command():'
        self.logger.debug(funcname)
        command = commandpacket(command=command, device_uuid=self.uuid, thread_uuid=self.thread_uuid,devicename=self.name,host=self.redvypr.hostinfo,devicemodulename=self.devicemodulename)
        # TODO, this should be done by commandpacket
        if(data is not None):
            if type(data) == dict:
                command.update(data)
            else:
                raise TypeError('data needs to be a dictionary')

        if(self.thread_running()):
            #print('Sending command',command)
            self.__send_command__(command)
        else:
            self.logger.warning(funcname + ' thread is not running, doing nothing')

    def __check_thread_status(self):
        """
        Regular thread status thread to test of thread is already stopped. Emit signal if stopped
        :return:
        """
        self.__stop_checks -= 1
        running2 = self.thread_running()
        info_dict = {}
        info_dict['uuid'] = self.uuid
        info_dict['thread_running'] = running2
        # Thread is not running anymore, stop timer and sent signal
        if info_dict['thread_running'] == False:
            self.__stoptimer__.stop()
            self.thread_stopped.emit(info_dict)
        if self.__stop_checks < 0:
            self.logger.warning('Could not stop device, giving up')
            self.__stoptimer__.stop()

    def thread_stop(self):
        """
        Sends a stop command to the thread_communication queue
        Returns:

        """
        funcname = __name__ + '.thread_stop():'
        self.logger.debug(funcname)
        command = commandpacket(command='stop', device_uuid=self.uuid,thread_uuid=self.thread_uuid)
        #print('Sending command',command)
        running = self.thread_running()
        if(running):
            self.__stoptimer__.stop()
            self.__stop_checks = 10 # Number of checks of stopped
            self.__send_command__(command)
            # Start a timer and check if the thread was stopped
            self.__stoptimer__.start(250)
        else:
            self.logger.warning(funcname + ' thread is not running, doing nothing')

    def thread_start(self, config=None):
        """ Starts the device thread, it calls the self.start function with the arguments

        start(self, device_info, config, dataqueue, datainqueue, statusqueue)

        config is a deepcopy of self.config: config = copy.deepcopy(self.config). The deepcopy converts the
        configuration dictionary into an ordinary dictionary "forgetting" additional information and making the
        configuration hashable as it is needed for a multiprocess.

        Args:
            config [default=None]: confuguration dictionary, if none a deepcopy of self.config is used, otherwise config


        """
        funcname = __name__ + '.thread_start():'

        #logger.debug(funcname + 'Starting device: ' + str(device.name))
        self.logger.debug(funcname + 'Starting device: ' + str(self.name))
        sendict = {}#self.__sendict__
        # Find the right thread to start
        if True:
            if True:
                running = self.thread_running()
                if(running):
                    self.logger.warning(funcname + ':thread/process is already running, doing nothing')
                else:
                    try:
                        # The arguments for the start function
                        thread_uuid = 'thread_' + str(uuid.uuid1())
                        device_info = {'device':self.name,'uuid':self.uuid,'thread_uuid':thread_uuid,'hostinfo':self.redvypr.hostinfo,'address_str':self.address_str}
                        if config is None:
                            self.logger.debug('Using internal configuration')
                            self.logger.debug('Pydantic configuration')
                            try:
                                config = self.custom_config.model_dump()
                            except:
                                raise ValueError('Could not dump custom model config')
                        else:
                            self.logger.debug('Using external configuration')

                        if self.device_parameter.clear_datainqueue_before_thread_starts:
                            while self.datainqueue.empty() == False:
                                try:
                                    data = self.datainqueue.get(block=False)
                                except:
                                    break
                        args = (device_info, config, self.dataqueue, self.datainqueue, self.statusqueue)
                        if True:
                            self.thread = deviceQThread(startfunction=self.start,start_arguments=args)
                        elif self.mp == 'thread':
                            self.logger.info(funcname + 'Starting as thread')
                            self.thread = threading.Thread(target=self.start, args=args, daemon=True)
                        else:
                            self.logger.info(funcname + 'Starting as process')
                            self.thread = multiprocessing.Process(target=self.start, args=args)

                            #self.logger.info(funcname + 'started {:s} as process with PID {:d}'.format(self.name, self.thread.pid))

                        self.thread.start()

                        sendict['thread'] = self.thread
                        self.logger.info(funcname + 'started {:s}'.format(self.name))
                        # Update the device and the devicewidgets about the thread status
                        running2 = self.thread_running()


                        try: # If the device has a thread_status function
                            self.thread_status({'threadalive':running2})
                        except:
                            pass

                        self.thread_uuid = thread_uuid
                        # Sending metadata
                        compacket = self.redvypr.get_metadata_commandpacket()
                        for addr in self.subscribed_addresses:
                            if compacket in addr:
                                self.datainqueue.put(compacket)

                        info_dict = {}
                        info_dict['uuid'] = self.uuid
                        info_dict['thread_uuid'] = thread_uuid
                        info_dict['thread_running'] = running2
                        self.thread_started.emit(info_dict)  # Notify about the started thread

                        return self.thread

                    except Exception:
                        self.logger.warning(funcname + 'Could not start thread.',exc_info=True)

                        return None

    def __str__(self):
        return self.description
    
    def finalize_init(self):
        """
        Dummy function, can be defined by the user. Function is called after the device is added to redvypr.

        Returns:

        """
        pass


    def get_hosts(self):
        """
        Returns:
            List of redvypr host instances this device has been seen
        """
        hosts = []
        devs = self.get_deviceaddresses()
        for d in devs:
            hosts.append(d.hostname)

        hosts = list(set(hosts))
        hosts.sort()
        return hosts


    def get_deviceaddresses(self, local=None):
        """
        Returns a list with RedvyprAddresses of all data devices that publish data via this device. This is in many cases
        the device itself but can also forwarded devices (i.e. iored) or because the device publishes data with different

        Args:
            local: True for local devices only, False for remote devices only, None for both

        Returns:
            List of redvypr_address

        """
        addr_str = list(self.statistics['device_redvypr'].keys())
        addr_list = []
        for a in addr_str:
            raddr = RedvyprAddress(a)
            if local is None:
                addr_list.append(raddr)
            else: # Check if we have a local or remote address
                raddr_local = raddr.uuid == self.host_uuid
                if local == raddr_local:
                    addr_list.append(raddr)

        return addr_list

    def get_datakeys(self, local=None):
        """
        Returns a list of all datakeys this device is providing
        Returns:
            List of datakeys (str)
        """
        devaddrs = self.get_deviceaddresses(local)
        datakeys = []
        for devaddr in devaddrs:
            dkeys = self.statistics['device_redvypr'][devaddr.address_str]['datakeys']
            datakeys.extend(dkeys)

        # Sort the datakeys and make them unique
        datakeys = list(set(datakeys))
        datakeys.sort()
        return datakeys

    def get_datastreams(self,local=None):
        """
        Returns a list of all datastreams this device is providing
        Returns:
            List of datastreams (str)
        """
        devaddrs = self.get_deviceaddresses(local)
        datastreams = []
        for devaddr in devaddrs:
            dkeys = self.statistics['device_redvypr'][devaddr.address_str]['datakeys']
            for dkey in dkeys:
                raddr = RedvyprAddress(devaddr, datakey=dkey)
                dstr = raddr.get_str()
                datastreams.append(dstr)


        return datastreams

    def get_packetids(self):
        """
        Returns a list of all packetids this device has seen
        Returns:
            List of packetids (str)
        """
        devaddrs = self.get_deviceaddresses()
        packetids = []
        for devaddr in devaddrs:
            if len(devaddr.packetid)>0:
                packetids.append(devaddr.packetid)

        # Sort the datakeys and make them unique
        packetids = list(set(packetids))
        packetids.sort()
        return packetids

    def get_device_info(self, address=None):
        """
        Returns a deepcopy that is saved in self.statistics['device_redvypr'] containing information about all devices
        that have been published through this device.
        Note: name was def get_data_provider_info(self):

        Returns:
            Dictionary with the device address (str) as key

        """
        if address is None:
            d = copy.deepcopy(self.statistics['device_redvypr'])
            return d
        else:
            if type(address) == str:
                raddr = RedvyprAddress(address)
                dtmp = copy.deepcopy(self.statistics['device_redvypr'])
                for a in dtmp.keys():
                    if a in raddr:
                        d = dtmp[a]
                        return d


    def publishing_to(self):
        """

        Returns:
            List of redvypr_device this device is publishing to
        """
        funcname = __name__ + '.publishing_to()'
        #self.logger.debug(funcname)
        devs = self.redvypr.get_device_objects()
        devs_publishing_to = []
        for dev in reversed(devs):
            for subaddr in dev.subscribed_addresses:
                daddr = RedvyprAddress(subaddr)
                if (self.address in daddr) and (dev is not self):
                    devs_publishing_to.append(dev)
                    break

        return devs_publishing_to

    def get_subscribed_datastreams(self):
        """

        Returns:
            All datastreams of the subscribed devices

        """
        funcname = __name__ + '.get_subscribed_datastreams()'
        datastreams = []
        devs = self.redvypr.get_device_objects()
        for subaddr in self.subscribed_addresses:
            for dev in reversed(devs):
                datastreams_dev = dev.get_datastreams()
                for d in datastreams_dev:
                    daddr = RedvyprAddress(d)
                    if (subaddr in daddr) and (dev is not self):
                        datastreams.append(d)

        return datastreams

    def get_subscribed_devices(self):
        """
        Returns all redvypr_device this device is subscribed to.

        Returns:
             List with redvypr_device

        """
        funcname = __name__ + '.get_subscribed_devices()'
        #self.logger.debug(funcname)
        devs = self.redvypr.get_device_objects()
        devs_subscribed = []
        for subaddr in self.subscribed_addresses:
            for dev in reversed(devs):
                if (subaddr in dev.address) and (dev is not self):
                    devs_subscribed.append(dev)
                    devs.remove(dev)

        return devs_subscribed

    def get_subscribed_deviceaddresses(self):
        """
        List of redvypr devices addresses this device has subscribed. This is different from self.subcribed_addresses as it
        returns the existing devices that provide data to this device, in self.subscribed_addresses also regular expressions can exist.

        Returns:
            List of redvypr_address

        """
        funcname = __name__ + '.get_subscribed_deviceaddresses()'
        subaddresses = []
        raddresses = self.redvypr.get_deviceaddresses()

        #print(' A raddresses', raddresses)
        #print(' B subscribed addresses',self.subscribed_addresses)
        for subaddr in self.subscribed_addresses:
            for addr in reversed(raddresses):
                if subaddr in addr:
                    subaddresses.append(addr)
                    raddresses.remove(addr)

        #print('subaddresses', subaddresses)
        return subaddresses
            
    def unsubscribe_all(self):
        """
        """
        funcname = self.__class__.__name__ + '.unsubscribe_all()'
        self.logger.debug(funcname)
        self.subscribed_addresses = []

    def get_metadata_datakey(self, address, all_entries=True):
        """
        Returns the metadata of the redvypr address
        """
        funcname = self.__class__.__name__ + '.get_metadata_datakey():'
        self.logger.debug(funcname)
        if isinstance(address,str):
            daddr = redvypr.RedvyprAddress(address)
        else:
            daddr = address

        # d = copy.deepcopy(self.statistics['device_redvypr'])
        devinfo_all = copy.deepcopy(self.redvypr.deviceinfo_all)
        # print('Datastream',datastream,daddr)
        datakeyinfo = {}
        for hostdevice in devinfo_all:
            d = devinfo_all[hostdevice]
            for device in d:
                # Check first if an ordinary datakey is given, or an eval string to access subunits
                if daddr.datakeyeval == False: # Standard datakey
                    for dkey in d[device]['_keyinfo'].keys():
                        dstreamaddr_info = redvypr.RedvyprAddress(device, datakey=dkey)
                        # print('dstreamddr_info',dstreamaddr_info)
                        if daddr in dstreamaddr_info:
                            try:
                                datakeyinfo[dstreamaddr_info.get_str()].update(d[device]['_keyinfo'][dkey])
                            except:
                                datakeyinfo[dstreamaddr_info.get_str()] = d[device]['_keyinfo'][dkey]
                else: # Eval string, here things have to be done manually.
                    self.logger.debug(funcname + 'Found an eval string, searching for matching patterns')
                    # Split the evalstring into pieces separated by [] brackets and remove them subsequently until something was found
                    datakey_pieces = re.findall(r"\[.*?\]", daddr.datakey)
                    datakey_pieces_cumsum = []
                    for ipiece,piece in enumerate(datakey_pieces):
                        dtmp = ''
                        for ipiece2 in range(0,ipiece+1):
                            dtmp += datakey_pieces[ipiece2]

                        datakey_pieces_cumsum.append(dtmp)

                    datakey_pieces_cumsum.reverse()
                    #print('datakey pieces', datakey_pieces)
                    #print('datakey pieces cumsum', datakey_pieces_cumsum)
                    #for dkey in d[device]['_keyinfo'].keys():
                    if True:
                        #print('Daddr',daddr,daddr.datakey)
                        #print('key',dkey,daddr in dstreamaddr_info)
                        #print('Keyinfo',d[device]['_keyinfo'])
                        for keyeval in datakey_pieces_cumsum:
                            evalstr = '''d[device]['_keyinfo']''' + keyeval
                            #print('Evalstr',evalstr)
                            #print(d[device]['_keyinfo'])
                            try:
                                keyinfo = eval(evalstr,None)
                                dstreamaddr_info = redvypr.RedvyprAddress(device, datakey=keyeval)
                                #print('dstreamddr_info', dstreamaddr_info)
                                #print('Found something',keyinfo)
                                try:
                                    datakeyinfo[dstreamaddr_info.get_str()].update(keyinfo)
                                except:
                                    datakeyinfo[dstreamaddr_info.get_str()] = keyinfo
                                break
                            except:
                                #self.logger.info('Eval did not work',exc_info=True)
                                keyinfo = None

                        #print('Keyinfo',keyinfo)

        # Either return the first entry or all
        if all_entries == False:
            k = list(datakeyinfo.keys())
            if len(k) == 0:
                datakeyinfo = None
            else:
                datakeyinfo = datakeyinfo[k[0]]

            #print('Returning for address', address)
            #print('Returning datakeyinfo',datakeyinfo)
            return datakeyinfo
        else:
            return datakeyinfo

    def get_datakeyinfo_legacy(self,datastream):
        """
        Returns the datakeyinfo for the datastream

        Args:
            datastream:
            first_match:

        Returns:

        """
        funcname = self.__class__.__name__ + '.get_datakeyinfo()'
        self.logger.debug(funcname)
        daddr = redvypr.RedvyprAddress(datastream)
        #d = copy.deepcopy(self.statistics['device_redvypr'])
        devinfo_all = copy.deepcopy(self.redvypr.deviceinfo_all)
        #print('Datastream',datastream,daddr)
        datakeyinfo = {}
        for hostdevice in devinfo_all:
            d = devinfo_all[hostdevice]
            for device in d:
                for dkey in d[device]['_keyinfo'].keys():
                    dstreamaddr_info = redvypr.RedvyprAddress(device, datakey = dkey)
                    #print('dstreamddr_info',dstreamaddr_info)
                    if daddr in dstreamaddr_info:
                        try:
                            datakeyinfo[dstreamaddr_info.get_str()].update(d[device]['_keyinfo'][dkey])
                        except:
                            datakeyinfo[dstreamaddr_info.get_str()] = d[device]['_keyinfo'][dkey]

        return datakeyinfo

    def get_datastream_keyinfo_legacy(self, datastream):
        """
        Returns the datakeyinfo for the datastream

        Args:
            datastream:
            first_match:

        Returns:

        """
        funcname = self.__class__.__name__ + '.get_datastream_keyinfo()'
        self.logger.debug(funcname)
        daddr = redvypr.RedvyprAddress(datastream)
        d = copy.deepcopy(self.statistics['device_redvypr'])
        # print('Datastream',datastream,daddr)
        for device in d:
            for dkey in d[device]['_keyinfo'].keys():
                dstreamaddr_info = redvypr.RedvyprAddress(device, datakey=dkey)
                # print('dstreamddr_info',dstreamaddr_info)
                if daddr in dstreamaddr_info:
                    return d[device]['_keyinfo'][dkey]

        return None

    def get_config(self):
        """
        Returns a RedvyprDeviceConfig of the device
        """
        funcname = __name__ + '.get_config():'
        self.logger.debug(funcname)
        # Treat subscriptions
        subscriptions = []
        for raddr in self.subscribed_addresses:
            subscriptions.append(raddr.address_str)
        base_config = RedvyprDeviceBaseConfig(**self.device_parameter.model_dump())
        try:
            custom_config_dict = self.custom_config.model_dump()
        except:
            self.logger.debug(funcname + 'Could not create a dump of custom config', exc_info=True)
            custom_config_dict = None
        config = RedvyprDeviceConfig(base_config=base_config, custom_config=custom_config_dict,
                                     devicemodulename=self.devicemodulename, subscriptions=subscriptions)

        self.logger.debug(funcname + 'Config: {}'.format(config))
        return config

    def get_status(self):
        statusstr = 'Status'
        return statusstr

    def get_info(self):
        """
        Returns a dictionary with the essential info of the device
        Returns:

        """
        info_dict = {}
        info_dict['uuid'] = self.uuid
        info_dict['name'] = self.name

        return info_dict

    def get_metadata(self, address,mode='merge'):
        """
        Gets the metadata of the redvypr address
        :param address:
        :return:
        """
        funcname = __name__ + '.get_metadata({},{}):'.format(str(address),str(mode))
        self.logger.debug(funcname)
        metadata = redvypr.packet_statistic.get_metadata(self.statistics,address,mode=mode)
        return metadata

    def set_metadata(self, address, metadata):
        """
        Sets the metadata of address.
        :param address:
        :param metadata:
        :return:
        """
        funcname = __name__ + '.set_metadata():'
        self.logger.debug(funcname)
        address_str = RedvyprAddress(address).address_str
        datapacket = redvypr.data_packets.commandpacket(command='reply')
        datapacket['_metadata'] = {}
        datapacket['_metadata'][address_str] = metadata
        self.dataqueue.put(datapacket)
        data = self.distribute_data_replyqueue.get()
        self.logger.debug(funcname + 'Metadata sent')

    def set_metadata_entry(self, address, metadata_key='unit', metadata_entry=None):
        """
        Convenience function to set one entry of the metadata of a redvypr address
        :param address:
        :param metadata_key:
        :param metadata_entry:
        :return:
        """
        funcname = __name__ + '.set_metadata_entry():'
        self.logger.debug(funcname)
        metadata = {}
        metadata[metadata_key:metadata_entry]
        self.set_metadata(address,metadata)

    def print_info(self):
        """ Displays information about the device

        """
        print('get_info():')
        print('Name' + self.name)
        print('redvypr' + self.redvypr)

    def __str__(self):
        return 'redvypr_device (' + self.devicemodulename + ') ' + self.address_string()




""" Utility to have the version in redvypr available
"""
import pkg_resources
import os
import sys
import logging

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('version')
logger.setLevel(logging.INFO)

#https://stackoverflow.com/questions/7674790/bundling-data-files-with-pyinstaller-onefile/13790741#13790741
def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)

# Get the version of redvypr
_version_file = pkg_resources.resource_filename('redvypr','VERSION')

# This is a workaround to read the VERSION file in a pyinstaller environment in linux (redvypr exectuable and redvypr directory cannot life together)
if(os.path.exists(_version_file)):
    pass
else:
    _version_file = resource_path('VERSION')
    if (os.path.exists(_version_file)):
        pass
    else: # pyinstaller windows10
        logger.warning('Could not load version file {}'.format(_version_file))

logger.debug('Opening version file {}'.format(_version_file))
with open(_version_file) as _version_f:
   version = _version_f.read().strip()

_version_f.close()



"""

influxdb device

Configuration options for a network device:

.. code-block::

"""


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import threading
import copy
import zmq
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
import influxdb_client


description = 'Device connects to an influxdb database'

config_template = {}
config_template['bucket']      = {'type': 'str','default':'redvypr'}
config_template['url']         = {'type': 'str','default':"http://localhost:8086"}
config_template['token']       = {'type': 'str','default':''}
config_template['datastreams'] = {'type': 'list'}
config_template['redvypr_device'] = {}
config_template['redvypr_device']['publish']     = True
config_template['redvypr_device']['subscribe']   = False
config_template['redvypr_device']['description'] = description
redvypr_devicemodule = True


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('test_device')
logger.setLevel(logging.DEBUG)



def start(device_info,config=None,dataqueue=None,datainqueue=None,statusqueue=None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    print('config',config)
    i = 0
    while True:
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if(data is not None):
            command = check_for_command(data,thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break

        dstr = config['string_send']
        print(dstr)
        dataqueue.put(dstr)
        dataqueue.put({'count':i})
        i+=1
        time.sleep(config['delay_s'])
        






"""
The serial device reads data from the serial interface and publishes them.
"""
import copy
import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import serial
import serial.tools.list_ports
import logging
import sys
import threading
import pydantic
import typing
import re
from redvypr.data_packets import check_for_command, create_datadict
from redvypr.redvypr_address import RedvyprAddress
from redvypr.device import RedvyprDevice
import redvypr.files as redvypr_files
from redvypr.data_packets import RedvyprMetadata, RedvyprDeviceMetadata
import redvypr.devices.plot.plot_widgets as redvypr_plot_widgets
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget, pydanticDeviceConfigWidget, dictQTreeWidget, datastreamMetadataWidget

_logo_file = redvypr_files.logo_file
_icon_file = redvypr_files.icon_file
description = 'Reading data from a serial device'


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('serial')
logger.setLevel(logging.INFO)


packet_start = ['<None>','$','custom']
packet_delimiter = ['None','CR/LF','LF','custom']
redvypr_devicemodule = True

class RedvyprSerialDeviceMetadata(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    comment: str = ''

class DeviceBaseConfig(pydantic.BaseModel):
    """
    BaseConfig
    Attributes:
        publishes: True
    """
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Reads to and writes from a serial devices'
    gui_tablabel_display: str = 'Serial device status'
    gui_icon: str = 'mdi.serial-port'

class SerialDeviceCustomConfig(pydantic.BaseModel):
    use_device: bool = pydantic.Field(default=True, description='Flag if the device should be used')
    baud: int = 9600
    parity: typing.Literal[serial.PARITY_NONE, serial.PARITY_ODD, serial.PARITY_EVEN, serial.PARITY_MARK, serial.PARITY_SPACE] = pydantic.Field(default=serial.PARITY_NONE)
    stopbits: typing.Literal[serial.STOPBITS_ONE, serial.STOPBITS_ONE_POINT_FIVE,serial.STOPBITS_TWO] = pydantic.Field(default=serial.STOPBITS_ONE)
    bytesize: typing.Literal[serial.EIGHTBITS, serial.SEVENBITS, serial.SIXBITS] = pydantic.Field(default=serial.EIGHTBITS)
    dt_poll: float = 0.05
    dt_maxwait: float = pydantic.Field(default=-1.0,description='Wait time in s for valid data, if time without a valid packets exceeds dt_maxwait the comport is closed and the read thread is stopped')
    chunksize: int = pydantic.Field(default=0, description='The maximum amount of bytes read with one chunk')
    packetdelimiter: str = pydantic.Field(default='LF', description='The delimiter to distinguish packets, leave empty to disable')
    packetstart: str = pydantic.Field(default='', description='The delimiter to distinguish packets, leave empty to disable')
    device: str = pydantic.Field(default='')
    devicename: str = pydantic.Field(default='')


class DeviceCustomConfig(pydantic.BaseModel):
    serial_devices: typing.Optional[typing.List[SerialDeviceCustomConfig]] = pydantic.Field(default=[])
    ignore_devices: str = pydantic.Field(default='.*/ttyS[0-9][0-9]*', description='Regular expression of serial device names, that are ignored.')
    baud_standard: list = [300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200]

def read_serial(device_info, config={}, dataqueue=None, datainqueue=None, statusqueue=None):
    """ Thread to read one serial device, is started for each device/comport

    """
    funcname = __name__ + '.read_serial()'
    logger.debug(funcname + ':Starting reading serial data')
    devicename_redvypr = device_info['device']
    chunksize = config['chunksize']  # The maximum amount of bytes read with one chunk
    serial_name = config['device']
    baud = config['baud']
    parity = config['parity']
    stopbits = config['stopbits']
    bytesize = config['bytesize']
    dt_poll = config['dt_poll']
    dt_maxwait = config['dt_maxwait']
    tnewpacket = time.time() # Time a new packet has arrived

    logger.debug('Starting to read serial device {:s} with baudrate {:d}'.format(serial_name,baud))

    devicename = config['devicename']
    # Get the packet end and packet start characters
    newpacket = config['packetdelimiter']
    # Check if a delimiter shall be used (\n, \r\n, etc ...)
    if (len(newpacket) > 0):
        FLAG_DELIMITER = True
    else:
        FLAG_DELIMITER = False
    if (type(newpacket) is not bytes):
        newpacket = newpacket.encode('utf-8')


    startpacket = config['packetstart'].encode('utf-8')

    rawdata_all = b''
    dt_update = 1  # Update interval in seconds
    bytes_read = 0
    sentences_read = 0
    bytes_read_old = 0  # To calculate the amount of bytes read per second
    t_update = time.time()
    serial_device = False
    if True:
        try:
            serial_device = serial.Serial(serial_name, baud, parity=parity, stopbits=stopbits, bytesize=bytesize,
                                          timeout=0)
            data = create_datadict(device= devicename_redvypr,packetid=devicename)
            data['t'] = time.time()
            data['comport'] = serial_device.name
            data['status'] = 'reading'
            # statusqueue.put(data)
            dataqueue.put(data)
            # print('Serial device 0',serial_device)
            # serial_device.timeout(0.05)
            # print('Serial device 1',serial_device)
        except Exception as e:
            # print('Serial device 2',serial_device)
            logger.debug(funcname + ': Exception open_serial_device {:s} {:d}: '.format(serial_name, baud) + str(e))
            data = create_datadict(packetid=devicename)
            data['t'] = time.time()
            data['comport'] = serial_device.name
            data['status'] = 'could not open'
            # statusqueue.put(data)
            dataqueue.put(data)
            return False

    got_dollar = False
    while True:
        # Note: Here commands could be sent as well
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if (data is not None):
            command = data
            # logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                if command == 'stop':
                    serial_device.close()
                    sstr = funcname + ': Command is for me ({:s}): {:s}'.format(config['device'],str(command))
                    logger.debug(sstr)
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    return

        time.sleep(dt_poll)
        ndata = serial_device.inWaiting()
        try:
            rawdata_tmp = serial_device.read(ndata)
        except Exception as e:
            print(e)
            # print('rawdata_tmp', rawdata_tmp)

        nread = len(rawdata_tmp)
        if True:
            if nread > 0:
                bytes_read += nread
                rawdata_all += rawdata_tmp
                #print('rawdata_all',rawdata_all)
                FLAG_CHUNK = (len(rawdata_all)) > chunksize and (chunksize > 0)
                if (FLAG_CHUNK):
                    data = create_datadict(device=devicename_redvypr,packetid=devicename)
                    data['t'] =  time.time()
                    data['data'] = rawdata_all
                    data['comport'] = serial_device.name
                    data['bytes_read'] = bytes_read
                    dataqueue.put(data)
                    rawdata_all = b''

                # Check if the newpacket character in the data
                if (FLAG_DELIMITER):
                    FLAG_CHAR = newpacket in rawdata_all
                    if (FLAG_CHAR):
                        rawdata_split = rawdata_all.split(newpacket)
                        #print('Found packets')
                        # print('rawdata_all', rawdata_all)
                        if (len(rawdata_split) > 1):  # If len==0 then character was not found
                            for ind in range( len(rawdata_split) - 1):  # The last packet does not have the split character
                                if rawdata_split[ind].startswith(startpacket):
                                    #print('Startmatch')
                                    tnewpacket = time.time()
                                    sentences_read += 1
                                    raw = rawdata_split[ind] + newpacket  # reconstruct the data
                                    # print('raw', raw)
                                    data = create_datadict(device=devicename_redvypr,packetid=devicename)
                                    data['t' ] = tnewpacket
                                    data['data'] = raw
                                    data['comport'] = serial_device.name
                                    data['bytes_read'] = bytes_read
                                    data['sentences_read'] = sentences_read
                                    #print('data',data)
                                    dataqueue.put(data)

                            rawdata_all = rawdata_split[-1]

        if ((time.time() - tnewpacket) > dt_maxwait) and (dt_maxwait > 0):
            logger.warning('Did not find valid packet on serial device {:s}'.format(serial_name))
            data = create_datadict(device=devicename_redvypr,packetid=devicename)
            data['t'] = time.time()
            data['comport'] = serial_device.name
            data['status'] = 'timout (dt_maxwait)'
            #statusqueue.put(data)
            dataqueue.put(data)
            return

        if ((time.time() - t_update) > dt_update):
            dbytes = bytes_read - bytes_read_old
            bytes_read_old = bytes_read
            bps = dbytes / dt_update  # bytes per second
            data = create_datadict(device=devicename_redvypr,packetid=devicename)
            data['t'] = time.time()
            data['comport'] = serial_device.name
            data['bps'] = bps
            data['bytes_read'] = bytes_read
            dataqueue.put(data)
            #statusqueue.put(data)
            # print('ndata',len(rawdata_all),'rawdata',rawdata_all,type(rawdata_all))
            # print('bps',bps)
            t_update = time.time()

def start(device_info, config={}, dataqueue=None, datainqueue=None, statusqueue=None):
    """

    """
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Starting reading serial data')
    logger.debug('Will open comports {:s}'.format(str(config['serial_devices'])))

    serial_threads = []
    serial_threads_datainqueues = []
    dt_poll = 0.05
    for comportconfig in config['serial_devices']:
        logger.debug('Processing device {}'.format(comportconfig['devicename']))
        if comportconfig['use_device'] == False:
            logger.debug('Ignoring device {}'.format(comportconfig['devicename']))
        else:
            logger.debug('Configuring device {}'.format(comportconfig['devicename']))
            queuesize = 100
            config_thread = copy.deepcopy(comportconfig)
            datainqueue_thread = queue.Queue(maxsize=queuesize)
            serial_threads_datainqueues.append(datainqueue_thread)
            args = [device_info,config_thread,dataqueue,datainqueue_thread,statusqueue]
            serial_thread = threading.Thread(target=read_serial, args=args, daemon=True)
            serial_threads.append(serial_thread)
            logger.debug('Starting thread with config: {:s}'.format(str(config_thread)))
            serial_thread.start()

    while True:
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if (data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            # logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                if command == 'stop':
                    sstr = funcname + ': Command is for me: {:s}'.format(str(command))
                    logger.debug(sstr)
                    # Sending stop command to the threads
                    for datainqueue in serial_threads_datainqueues:
                        datainqueue.put('stop')

                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    return

        # Check if any of the threads is running, if not stop main thread as well
        all_dead = True
        for serial_thread in serial_threads:
            if serial_thread.is_alive():
                all_dead = False
                break

        if all_dead:
            return

        time.sleep(dt_poll)





class Device(RedvyprDevice):
    """
    heatflow_serial device
    """

    def __init__(self, **kwargs):
        """
        """
        funcname = __name__ + '__init__()'
        super(Device, self).__init__(**kwargs)
        logger.debug(funcname)
        self.get_comports()

    def thread_start(self):
        try:
            self.deviceinitwidget.widgets_to_config()
        except:
            logger.debug('Could not init using gui settings',exc_info=True)
        super().thread_start()

    def get_comports(self):
        funcname = __name__ + '.get_comports():'
        logger.debug(funcname)
        comports = serial.tools.list_ports.comports()
        self.comports = []
        serial_devices_new = []
        for comport in comports:
            FLAG_DEVICE_EXISTS = False
            for d in self.custom_config.serial_devices:
                if comport.device == d.device:
                    logger.debug(funcname + ' Found new device {}'.format(d.device))
                    FLAG_DEVICE_EXISTS = True
                    serial_devices_new.append(d)
                    self.comports.append(comport)
                    break

            if re.match(self.custom_config.ignore_devices,comport.device):
                logger.info('Ignoring device {}'.format(comport.device))
            else:
                if FLAG_DEVICE_EXISTS == False:
                    devicename = comport.device.split('/')[-1]
                    config = SerialDeviceCustomConfig(device=comport.device, devicename=devicename)
                    #self.config.serial_devices.append(config)
                    serial_devices_new.append(config)
                    self.comports.append(comport)

        self.custom_config.serial_devices = serial_devices_new
        logger.debug(funcname + ' serial devices {}'.format(self.custom_config.serial_devices))


class initDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QVBoxLayout(self)
        self.device   = device
        #self.serialwidget = QtWidgets.QWidget()
        self.serialwidgets_parent = QtWidgets.QWidget()
        layout_all = QtWidgets.QGridLayout(self.serialwidgets_parent)
        self.layout_serialwidgets = layout_all
        self.serialwidgets = []
        self.label= QtWidgets.QLabel("Serial device")

        self.comscanbutton = QtWidgets.QPushButton("Rescan comports")
        self.comscanbutton.clicked.connect(self.comscan_clicked)
        # self.comscanbutton.setCheckable(True)
        layout.addWidget(self.comscanbutton)


        self.init_serialwidgets()
        layout.addWidget(self.serialwidgets_parent)

        self.startbutton = QtWidgets.QPushButton("Start")
        self.startbutton.clicked.connect(self.start_clicked)
        self.startbutton.setCheckable(True)
        layout.addWidget(self.startbutton)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

    def comscan_clicked(self):
        funcname = __name__ + '.comscan_clicked()'
        logger.debug(funcname)
        self.device.get_comports()
        layout = self.layout_serialwidgets
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)
        self.init_serialwidgets()
        # Redraw also the devicedisplaywidget
        self.device.devicedisplaywidget.populate_statustable()


    def init_serialwidgets(self):
        layout_all = self.layout_serialwidgets
        layout_all.addWidget(QtWidgets.QLabel('Comport'), 0, 0)
        layout_all.addWidget(QtWidgets.QLabel('Devicename'), 0, 1)
        layout_all.addWidget(QtWidgets.QLabel('Baud'), 0, 2)
        layout_all.addWidget(QtWidgets.QLabel('Parity'), 0, 3)
        layout_all.addWidget(QtWidgets.QLabel('Databits'), 0, 4)
        layout_all.addWidget(QtWidgets.QLabel('Stopbits'), 0, 5)
        #layout_all.addWidget(QtWidgets.QLabel('Packet start'), 0, 6)
        layout_all.addWidget(QtWidgets.QLabel('Packet delimiter'), 0, 7)
        layout_all.addWidget(QtWidgets.QLabel('Packet size'), 0, 8)
        layout_all.addWidget(QtWidgets.QLabel('Metadata'), 0, 9)
        layout_all.addWidget(QtWidgets.QLabel('Use port'), 0, 10)
        lwidth = 80 # width of the qlineedits

        for irow,serial_device in enumerate(self.device.custom_config.serial_devices):
            widget = QtWidgets.QWidget()
            serialwidgetdict = {}
            serialwidgetdict['widget'] = widget
            serialwidgetdict['serial_device'] = serial_device
            self.serialwidgets.append(serialwidgetdict)
            layout = QtWidgets.QGridLayout(widget)
            # Serial baud rates
            #baud = [300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 576000, 921600]
            baudrates = self.device.custom_config.baud_standard
            # find the index of the baudrate of the device, if its none add it to the list
            try:
                ibaud = baudrates.index(serial_device.baud)
            except:
                baudrates.append(serial_device.baud)
                baudrates.sort()
                ibaud = baudrates.index(serial_device.baud)

            serialwidgetdict['combo_serial_devices'] = QtWidgets.QLabel(serial_device.device)
            # self._combo_serial_devices.currentIndexChanged.connect(self._serial_device_changed)
            serialwidgetdict['combo_serial_baud'] = QtWidgets.QComboBox()
            for b in baudrates:
                serialwidgetdict['combo_serial_baud'].addItem(str(b))

            serialwidgetdict['combo_serial_baud'].setCurrentIndex(ibaud)
            # creating a line edit to give the user the choice to edit a baudrate
            edit = QtWidgets.QLineEdit(self)
            onlyInt = QtGui.QIntValidator()
            edit.setValidator(onlyInt)
            # setting line edit
            serialwidgetdict['combo_serial_baud'].setLineEdit(edit)

            devicename = serial_device.devicename
            serialwidgetdict['devicename'] = QtWidgets.QLineEdit()
            serialwidgetdict['devicename'].setText(devicename)

            serialwidgetdict['combo_parity'] = QtWidgets.QComboBox()
            serialwidgetdict['combo_parity'].addItem('None')
            serialwidgetdict['combo_parity'].addItem('Odd')
            serialwidgetdict['combo_parity'].addItem('Even')
            serialwidgetdict['combo_parity'].addItem('Mark')
            serialwidgetdict['combo_parity'].addItem('Space')

            serialwidgetdict['combo_stopbits'] = QtWidgets.QComboBox()
            serialwidgetdict['combo_stopbits'].addItem('1')
            serialwidgetdict['combo_stopbits'].addItem('1.5')
            serialwidgetdict['combo_stopbits'].addItem('2')

            serialwidgetdict['combo_databits'] = QtWidgets.QComboBox()
            serialwidgetdict['combo_databits'].addItem('8')
            serialwidgetdict['combo_databits'].addItem('7')
            serialwidgetdict['combo_databits'].addItem('6')
            serialwidgetdict['combo_databits'].addItem('5')

            serialwidgetdict['button_serial_openclose'] = QtWidgets.QPushButton('Use')
            serialwidgetdict['button_serial_openclose'].setCheckable(True)
            serialwidgetdict['button_serial_openclose'].setChecked(serial_device.use_device)
            serialwidgetdict['button_serial_openclose'].clicked.connect(self.use_device_clicked)

            serialwidgetdict['button_metadata'] = QtWidgets.QPushButton('Metadata')
            serialwidgetdict['button_metadata'].clicked.connect(self.__metadata_clicked)
            serialwidgetdict['button_metadata'].__index_serial = irow

            # Define packet delimiter
            packet_del = serial_device.packetdelimiter
            serialwidgetdict['packet_ident_lab'] = QtWidgets.QLabel('Packet identification')
            serialwidgetdict['packet_ident'] = QtWidgets.QComboBox()
            idel_index = -1
            for idel, d in enumerate(packet_delimiter):
                if packet_del == d:
                    idel_index = idel
                elif idel == len(packet_delimiter) - 1:
                    if len(packet_del) > 0 and (idel_index==-1):
                        d = packet_del
                        idel_index = idel

                serialwidgetdict['packet_ident'].addItem(d)

            if idel_index >=0:
                serialwidgetdict['packet_ident'].setCurrentIndex(idel_index)
            serialwidgetdict['packet_ident'].setEditable(True)

            #.currentTextChanged.connect(self.current_text_changed)
            serialwidgetdict['packet_ident_lineedit'] = serialwidgetdict['packet_ident'].lineEdit()
            serialwidgetdict['packet_ident_lineedit'].__combo__ = serialwidgetdict['packet_ident']
            serialwidgetdict['packet_ident_lineedit'].textEdited.connect(self.handle_delimiter_text_edited)

            #serialwidgetdict['packet_ident'].setText(packet_del)
            #serialwidgetdict['packet_ident'].setFixedWidth(lwidth)
            # Set the tooltip
            desc = serial_device.model_fields['packetdelimiter'].description
            serialwidgetdict['packet_ident'].setToolTip(desc)

            #packet_start = serial_device.packetstart
            ## Handle new line character
            #packet_start = packet_start.replace('\n','\\n')
            #serialwidgetdict['packet_start_lab'] = QtWidgets.QLabel('Packet start')
            #serialwidgetdict['packet_start'] = QtWidgets.QLineEdit()
            #serialwidgetdict['packet_start'].setText(packet_start)
            #serialwidgetdict['packet_start'].setFixedWidth(lwidth)
            # Set the tooltip
            #desc = serial_device.model_fields['packetstart'].description
            #serialwidgetdict['packet_start'].setToolTip(desc)
            # Max packetsize
            serialwidgetdict['packet_size_lab'] = QtWidgets.QLabel("Maximum packet size")
            onlyInt = QtGui.QIntValidator()
            packet_size = str(serial_device.chunksize)
            serialwidgetdict['packet_size'] = QtWidgets.QLineEdit()
            serialwidgetdict['packet_size'].setValidator(onlyInt)
            serialwidgetdict['packet_size'].setText(packet_size)
            serialwidgetdict['packet_size'].setFixedWidth(lwidth)
            # Set the tooltip
            desc = serial_device.model_fields['chunksize'].description
            serialwidgetdict['packet_size'].setToolTip(desc)

            #layout.addWidget(serialwidgetdict['packet_start'], 2, 1)
            #layout.addWidget(serialwidgetdict['packet_start_lab'], 2, 0)
            layout.addWidget(serialwidgetdict['packet_ident'], 2, 1+2)
            layout.addWidget(serialwidgetdict['packet_ident_lab'], 2, 0+2)
            layout.addWidget(serialwidgetdict['packet_size_lab'], 2, 2+2)
            layout.addWidget(serialwidgetdict['packet_size'], 2, 3+2)

            layout_all.addWidget(serialwidgetdict['combo_serial_devices'], irow+1, 0)
            layout_all.addWidget(serialwidgetdict['devicename'], irow + 1, 1)
            layout_all.addWidget(serialwidgetdict['combo_serial_baud'], irow+1, 2)
            layout_all.addWidget(serialwidgetdict['combo_parity'], irow+1, 3)
            layout_all.addWidget(serialwidgetdict['combo_databits'], irow+1, 4)
            layout_all.addWidget(serialwidgetdict['combo_stopbits'], irow+1, 5)
            #layout_all.addWidget(serialwidgetdict['packet_start'], irow + 1, 6)
            layout_all.addWidget(serialwidgetdict['packet_ident'], irow + 1, 7)
            layout_all.addWidget(serialwidgetdict['packet_size'], irow + 1, 8)
            layout_all.addWidget(serialwidgetdict['button_metadata'], irow+1, 9)
            layout_all.addWidget(serialwidgetdict['button_serial_openclose'], irow+1, 10)

            #self.statustimer = QtCore.QTimer()
            #self.statustimer.timeout.connect(self.update_buttons)
            #self.statustimer.start(500)

        layout_all.setRowStretch(layout_all.rowCount(), 1)

    def __metadata_clicked(self):
        funcname = __name__ + '.__metadata_clicked():'
        logger.debug(funcname)
        self.widgets_to_config()
        irow = self.sender().__index_serial
        serial_device = self.device.custom_config.serial_devices[irow]
        devicename = serial_device.devicename
        devicename_redvypr = self.device.name

        metadata_device = copy.deepcopy(self.device.statistics['metadata'])
        deviceAddress = RedvyprAddress(devicename=self.device.name,packetid=devicename)
        try:
            metadata_raw = metadata_device[deviceAddress.address_str]
        except:
            logger.info('Could not load metadata', exc_info=True)
            metadata_raw = {}

        metadata = RedvyprSerialDeviceMetadata(**metadata_raw)
        print('Metadata', metadata)

        self.__metadata_edit = metadata
        self.__metadata_address = deviceAddress
        self.metadata_config = pydanticConfigWidget(metadata, configname=deviceAddress.address_str)
        self.metadata_config.config_editing_done.connect(self.__metadata_config_apply)
        self.metadata_config.setWindowTitle('redvypr, Metadata for {}'.format(devicename))
        self.metadata_config.show()

    def __metadata_config_apply(self):
        funcname = __name__ + '__metadata_config_apply():'
        logger.debug(funcname)

        #print('Metadata new', self.__metadata_edit)
        metadata = self.__metadata_edit.model_dump()
        self.device.set_metadata(self.__metadata_address, metadata)

    def handle_delimiter_text_edited(self, text):

        comboBox = self.sender().__combo__
        line_edit = self.sender()
        current_index = comboBox.currentIndex()
        nitems = comboBox.count()
        # Check of items shall be edited?
        #print('Hallo',current_index,nitems)
        if current_index != (nitems -1):  # Check if last item
            original_text = comboBox.itemText(current_index)
            #print('Not editable',original_text)
            line_edit.setText(original_text)
        else: # editable item
            pass

    def use_device_clicked(self):
        button = self.sender()
        if button.isChecked():
            button.setText('Use')
        else:
            button.setText('Ignore')
    def update_buttons(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """

        status = self.device.get_thread_status()
        thread_status = status['thread_running']

        if(thread_status):
            self._button_serial_openclose.setText('Close')
            self._combo_serial_baud.setEnabled(False)
            self._combo_serial_devices.setEnabled(False)
        else:
            self._button_serial_openclose.setText('Open')
            self._combo_serial_baud.setEnabled(True)
            self._combo_serial_devices.setEnabled(True)
        

    def widgets_to_config(self):
        """
        Copies the widget data to the configuration
        """
        funcname = __name__ + '.widgets_to_config()'
        logger.debug(funcname)
        # print('hallo',len(self.serialwidgets))
        configs = []
        for w in self.serialwidgets:
            serial_device = w['serial_device']
            config = {}
            serial_name = str(w['combo_serial_devices'].text())
            serial_device.device = serial_name
            devicename = str(w['devicename'].text())
            serial_device.devicename = devicename
            serial_baud = int(w['combo_serial_baud'].currentText())
            serial_device.baud = serial_baud
            stopbits = w['combo_stopbits'].currentText()
            if (stopbits == '1'):
                serial_device.stopbits = serial.STOPBITS_ONE
            elif (stopbits == '1.5'):
                serial_device.stopbits = serial.STOPBITS_ONE_POINT_FIVE
            elif (stopbits == '2'):
                serial_device.stopbits = serial.STOPBITS_TWO

            databits = int(w['combo_databits'].currentText())
            serial_device.bytesize = databits

            parity = w['combo_parity'].currentText()
            if (parity == 'None'):
                serial_device.parity = serial.PARITY_NONE
            elif (parity == 'Even'):
                serial_device.parity = serial.PARITY_EVEN
            elif (parity == 'Odd'):
                serial_device.parity = serial.PARITY_ODD
            elif (parity == 'Mark'):
                serial_device.parity = serial.PARITY_MARK
            elif (parity == 'Space'):
                serial_device.parity = serial.PARITY_SPACE

            #
            serial_device.chunksize = int(w['packet_size'].text())
            serial_device.packetdelimiter = w['packet_ident'].currentText()#.replace('\\n', '\n').replace('\\r', '\r')
            serial_device.packetdelimiter = serial_device.packetdelimiter.replace('CR/LF','\r\n')
            serial_device.packetdelimiter = serial_device.packetdelimiter.replace('LF', '\n')
            serial_device.packetdelimiter = serial_device.packetdelimiter.replace('None', '')
            print(funcname + 'Test',w['packet_ident'].currentText())
            #serial_device.packetstart = w['packet_start'].text()
            if w['button_serial_openclose'].isChecked():
                serial_device.use_device = True
            else:
                serial_device.use_device = False

    def start_clicked(self):
        button = self.sender()
        if ('Start' in button.text()):
            self.widgets_to_config()
            self.device.thread_start()
            button.setText('Stop')
        else:
            self.stop_clicked()

    def update_buttons(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """

        status = self.device.get_thread_status()
        thread_status = status['thread_running']
        # Running
        if (thread_status):
            self.startbutton.setText('Stop')
            self.startbutton.setChecked(True)
            #for w in self.config_widgets:
            #    w.setEnabled(False)
        # Not running
        else:
            self.startbutton.setText('Start')
            #for w in self.config_widgets:
            #    w.setEnabled(True)

            # Check if an error occured and the startbutton
            if (self.startbutton.isChecked()):
                self.startbutton.setChecked(False)
            # self.conbtn.setEnabled(True)

    def stop_clicked(self):
        """

        Returns:

        """
        self.device.thread_stop()
        self.startbutton.setText('Closing')


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout = QtWidgets.QGridLayout(self)
        self.device = device
        self.serialwidgets = []
        self.serialwidgetsdict = {}
        self.comporttable = QtWidgets.QTableWidget()
        layout.addWidget(self.comporttable,0,0)
        self.device = device
        self.populate_statustable()


    def populate_statustable(self):
        funcname = __name__ + '.populate_statustable():'
        logger.debug(funcname)
        self.comporttable.clear()
        columns = ['Comport', 'Bytes read', 'Packets read', 'Status', 'Show rawdata']
        self.comporttable.setColumnCount(len(columns))
        self.comporttable.horizontalHeader().ResizeMode(self.comporttable.horizontalHeader().ResizeToContents)
        self.comporttable.setHorizontalHeaderLabels(columns)
        comports = self.device.comports
        self.comporttable.setRowCount(len(comports))
        self.comports = []
        for irow, comport in enumerate(comports):
            self.comports.append(comport.device)
            serialwidgetdict = {}
            serialwidgetdict['datawidget'] = QtWidgets.QPlainTextEdit()
            serialwidgetdict['datawidget'].setReadOnly(True)
            serialwidgetdict['datawidget'].setMaximumBlockCount(10000)
            serialwidgetdict['datawidget'].setWindowIcon(QtGui.QIcon(_icon_file))
            serialwidgetdict['datawidget'].setWindowTitle(comport.device)
            self.serialwidgets.append(serialwidgetdict)
            self.serialwidgetsdict[comport.device] = serialwidgetdict

            item = QtWidgets.QTableWidgetItem(comport.device)
            self.comporttable.setItem(irow, 0, item)
            item = QtWidgets.QTableWidgetItem('0')
            self.comporttable.setItem(irow, 1, item)
            item = QtWidgets.QTableWidgetItem('0')
            self.comporttable.setItem(irow, 2, item)
            # Status
            item = QtWidgets.QTableWidgetItem('closed')
            self.comporttable.setItem(irow, 3, item)

            button = QtWidgets.QPushButton('Show')
            button.clicked.connect(self.__showdata__)
            button.displaywidget = serialwidgetdict['datawidget']
            self.comporttable.setCellWidget(irow, 4, button)

        self.comporttable.resizeColumnsToContents()

    def __showdata__(self):
        """
        Opens the raw data widget
        """
        button = self.sender()

        button.displaywidget.show()
        button.displaywidget.setFocus()
        button.displaywidget.raise_()
        button.displaywidget.activateWindow()
        #try:
        #    isshowing = button.showing
        #except:
        #    isshowing = False#

        #if not(isshowing):
        #    button.setText('Hide')
        #    button.showing = True
        #    button.displaywidget.show()
        #else:
        #    button.showing = False
        #    button.setText('Show')
        #    button.displaywidget.hide()
    def update_data(self, data):
        #print('data',data)
        try:
            comport = data['comport']
            datawidget = self.serialwidgetsdict[comport]['datawidget']
        except:
            return

        try:
            status = data['status']
            index = self.comports.index(comport)
            itemstatus = QtWidgets.QTableWidgetItem(status)
            self.comporttable.setItem(index, 3, itemstatus)
        except:
            pass
        try:
            index = self.comports.index(comport)
            bstr = "{:d}".format(data['bytes_read'])
            lstr = "{:d}".format(data['sentences_read'])
            itemb = QtWidgets.QTableWidgetItem(bstr)
            items = QtWidgets.QTableWidgetItem(lstr)
            itemstatus = QtWidgets.QTableWidgetItem('open')
            self.comporttable.setItem(index, 1, itemb)
            self.comporttable.setItem(index, 2, items)
            self.comporttable.setItem(index, 3, itemstatus)

        except Exception as e:
            # logger.exception(e)
            pass

        try:
            datawidget.insertPlainText(str(data['data']))
            datawidget.insertPlainText('\n')
        except Exception as e:
            #logger.exception(e)
            pass
        
